"""UGens.sc

Development note for UGen classes
---------------------------------
Subclasses should not use __init__ to implement graph logic, interface
methods are _new1, _multi_new, _multi_new_list, _init_ugen, _init_outputs
(from MultiOutUGen). UGen instances are created internally with
_create_ugen_object. This is because ugens build a graph with multichannel
expansion and optimizations so they might not return an instace object or
an instance of the same type.
"""

import inspect
import operator
import logging

from ..base import classlibrary as clb
from ..base import utils as utl
from ..base import builtins as bi
from ..base import absobject as aob
from . import _specialindex as _si
from . import _graphparam as gpp
from . import _fmtrw as frw


clb.ClassLibrary.late_imports(__name__,
    ('sc3.base.main', '_libsc3'),
    ('sc3.synth.ugens.trig', 'trg'),
    ('sc3.synth.ugens.pan', 'pan'),
    ('sc3.synth.ugens.infougens', 'ifu'),
    ('sc3.synth.ugens.filter', 'flr'),
    ('sc3.synth.ugens.oscillators', 'ocl'),
    ('sc3.synth.ugens.testugens', 'tsu'),
    ('sc3.synth.ugens.line', 'lne'),
    ('sc3.synth.ugens.demand', 'dmd'),
    ('sc3.synth.ugens.poll', 'pll'),
)


__all__ = []


_logger = logging.getLogger(__name__)


class ChannelList(list, gpp.UGenSequence, aob.AbstractObject):
    '''List wrapper for multichannel expansion graph operations.'''

    def __init__(self, obj=None):
        if obj is None:
            super().__init__()
        elif isinstance(obj, (str, tuple)):
            super().__init__([obj])
        elif hasattr(obj, '__iter__'):
            super().__init__(obj)
        else:
            super().__init__([obj])
        super(gpp.UGenSequence, self).__init__(self)


    ### AbstractObject interface ###

    def _compose_unop(self, selector):
        return utl.list_unop(selector, self, type(self))

    def _compose_binop(self, selector, other):
        return utl.list_binop(selector, self, other, type(self))

    def _rcompose_binop(self, selector, other):
        return utl.list_binop(selector, other, self, type(self))

    def _compose_narop(self, selector, *args):
        return utl.list_narop(selector, self, *args, t=type(self))


    ### UGen convenience methods (keep in sync) ###

    def _multichannel_perform(self, selector, *args):
        ret = type(self)()
        for item in self:
            item = gpp.ugen_param(item)
            ret.append(getattr(item, selector)(*args))
        return ret

    def dup(self, n=2):
        return ChannelList([self] * n)

    def sum(self):  # Implemented by Collection with optional function.
        return utl.list_sum(self, type(self))

    def madd(self, mul=1.0, add=0.0):
        return type(self)(MulAdd.new(i, mul, add) for i in self)

    # in SequenceableCollection L1148.

    def range(self, lo=0.0, hi=1.0):
        return self._multichannel_perform('range', lo, hi)

    def exprange(self, lo=0.01, hi=1.0):
        return self._multichannel_perform('exprange', lo, hi)

    def curverange(self, lo=0.0, hi=1.0, curve=-4):
        return self._multichannel_perform('curverange', lo, hi, curve)

    def unipolar(self, mul=1):
        return self._multichannel_perform('unipolar', mul)

    def bipolar(self, mul=1):
        return self._multichannel_perform('bipolar', mul)

    def clip(self, lo=0.0, hi=1.0):
        return self._multichannel_perform('clip', lo, hi)

    def fold(self, lo=0.0, hi=1.0):
        return self._multichannel_perform('fold', lo, hi)

    def wrap(self, lo=0.0, hi=1.0):
        return self._multichannel_perform('wrap', lo, hi)

    def min_nyquist(self):
        return type(self)(bi.min(item, ifu.SampleRate.ir * 0.5) for item in self)

    # degrad implemented with performUnaryOp, is not overridden here
    # raddeg implemented with performUnaryOp, is not overridden here

    def blend(self, other, frac=0.5):
        return self._multichannel_perform('blend', other, frac)

    def lag(self, time=0.1):
        return self._multichannel_perform('lag', time)

    def lag2(self, time=0.1):
        return self._multichannel_perform('lag2', time)

    def lag3(self, time=0.1):
        return self._multichannel_perform('lag3', time)

    def lagud(self, utime=0.1, dtime=0.1):
        return self._multichannel_perform('lagud', utime, dtime)

    def lag2ud(self, utime=0.1, dtime=0.1):
        return self._multichannel_perform('lag2ud', utime, dtime)

    def lag3ud(self, utime=0.1, dtime=0.1):
        return self._multichannel_perform('lag3ud', utime, dtime)

    def varlag(self, time=0.1, curvature=0, wrap=5, start=None):
        return self._multichannel_perform('varlag', time, curvature, wrap, start)

    def slew(self, up=1, down=1):
        return self._multichannel_perform('slew', up, down)

    def prune(self, min, max, type='minmax'):
        return self._multichannel_perform('prune', min, max, type)

    # snap is not implemented
    # softround is not implemented

    def linlin(self, inmin, inmax, outmin, outmax, clip='minmax'):
        return self._multichannel_perform('linlin', inmin, inmax, outmin,
                                          outmax, clip)

    def linexp(self, inmin, inmax, outmin, outmax, clip='minmax'):
        return self._multichannel_perform('linexp', inmin, inmax, outmin,
                                          outmax, clip)

    def explin(self, inmin, inmax, outmin, outmax, clip='minmax'):
        return self._multichannel_perform('explin', inmin, inmax, outmin,
                                          outmax, clip)

    def expexp(self, inmin, inmax, outmin, outmax, clip='minmax'):
        return self._multichannel_perform('expexp', inmin, inmax, outmin,
                                          outmax, clip)

    def lincurve(self, inmin, inmax, outmin, outmax, curve=-4, clip='minmax'):
        return self._multichannel_perform('lincurve', inmin, inmax, outmin,
                                          outmax, curve, clip)

    def curvelin(self, inmin, inmax, outmin, outmax, curve=-4, clip='minmax'):
        return self._multichannel_perform('curvelin', inmin, inmax, outmin,
                                          outmax, curve, clip)

    def bilin(self, incenter, inmin, inmax, outcenter, outmin, outmax,
              clip='minmax'):
        return self._multichannel_perform('bilin', incenter, inmin, inmax,
                                          outcenter, outmin, outmax, clip)

    def biexp(self, incenter, inmin, inmax, outcenter, outmin, outmax,
              clip='minmax'):
        return self._multichannel_perform('biexp', incenter, inmin, inmax,
                                          outcenter, outmin, outmax, clip)

    def moddif(self, that=0.0, mod=1.0):
        return self._multichannel_perform('moddif', that, mod)

    # in Array.sc

    # num_channels, no (is len, UGen don't really know about channels), TODO: ensure consistency.
    # source_ugen, no if not needed.

    # Synth debug

    def poll(self, trig=10, label=None, trig_id=-1):
        if label is None:
            label = [f'ChannelList UGen [{i}]' for i in range(len(self))]
        return pll.Poll.new(trig, self, label, trig_id)

    def dpoll(self, label=None, run=1, trig_id=-1):
        if label is None:
            label = [f'ChannelList UGen [{i}]' for i in range(len(self))]
        return dmd.Dpoll(self, label, run, trig_id)

    def check_bad_values(self, id=0, post=2):
        return self._multichannel_perform('check_bad_values', id, post)


    ### Override list methods ###

    def __add__(self, other): # +
        return self._compose_binop(operator.add, other)

    def __iadd__(self, other): # +=
        return self._compose_binop(operator.add, other)

    def __mul__(self, other): # *
        return self._compose_binop(operator.mul, other)

    def __rmul__(self, other):
        return self._rcompose_binop(operator.mul, other)

    def __imul__(self, other): # *=
        return self._compose_binop(operator.mul, other)

    def __lt__(self, other): # <
        return self._compose_binop(operator.lt, other)

    def __le__(self, other): # <=
        return self._compose_binop(operator.le, other)

    # def __eq__(self, other):
    #     return self._compose_binop(operator.eq, other)

    # def __ne__(self, other):
    #     return self._compose_binop(operator.ne, other)

    def __gt__(self, other): # >
        return self._compose_binop(operator.gt, other)

    def __ge__(self, other): # >=
        return self._compose_binop(operator.ge, other)


    def __repr__(self):
        return f'ChannelList({super().__repr__()})'


class MetaSynthObject(type):
    # NOTE: Do not use default rate within the library.
    def __call__(cls, *args, **kwargs):
        if 'urate' in kwargs:
            rate = kwargs.pop('urate')
        else:
            rate = cls._default_rate
        rate = cls._method_selector_for_rate(rate)
        return getattr(cls, rate)(*args, **kwargs)


class SynthObject(gpp.UGenParameter, metaclass=MetaSynthObject):
    # NOTE: Sum3 and Sum4 don't define rate before calling _multi_new_list.
    _valid_rates = {'audio', 'control', 'demand', 'scalar', None}
    _default_rate = 'audio'

    @classmethod
    def _create_ugen_object(cls, rate):
        obj = cls.__new__(cls)
        super(SynthObject, obj).__init__(obj)
        obj._rate = rate
        obj._inputs = ()

        obj._synthdef = None  # Is_current_synthdef after _add_to_synth.
        obj._channels = []  # For MultiOutUGen, related to _synth_index.
        obj._synth_index = -1  # Order in built graph.
        obj._output_index = 0  # Used by OutputProxy.
        obj._special_index = 0  # Server op index.

        # Topo sort.
        obj._antecedents = None  # set()  # _init_topo_sort
        obj._descendants = None  # set()  # _init_topo_sort
        obj._width_first_antecedents = None  # list()  # _width_first_ugens[:].

        return obj

    @property
    def rate(self):
        return self._rate

    @property
    def inputs(self):
        return self._inputs

    @classmethod
    def _new1(cls, rate, *args):
        '''
        This method returns a single instance of the UGen, not multichannel
        expanded. It is called inside _multi_new_list, whenever a new single
        instance is needed.
        '''
        obj = cls._create_ugen_object(rate)
        obj._add_to_synth()
        return obj._init_ugen(*args)

    @classmethod
    def _multi_new(cls, *args):
        return cls._multi_new_list(list(args))

    @classmethod
    def _multi_new_list(cls, args):
        '''
        These methods are responsible for multichannel expansion. They call
        UGen._new1(rate, *args) for each parallel combination. Most UGen.ar/kr
        methods delegate to UGen.multiNewList. The first argument is rate, then
        the rest of the arguments as in UGen._new1(rate, *args).
        '''

        # Single channel, one ugen.
        length = 0
        args = gpp.ugen_param(args)._as_ugen_input(cls)
        for item in args:
            if isinstance(item, list):
                length = max(length, len(item))
        if length == 0:
            cls._check_valid_rate_name(args[0])
            return cls._new1(*args)

        # Multichannel expansion.
        new_args = [None] * len(args)
        results = [None] * length
        for i in range(length):
            for j, item in enumerate(args):
                new_args[j] = (
                    item[i % len(item)] if isinstance(item, list) else item)
            cls._check_valid_rate_name(new_args[0])
            results[i] = cls._multi_new(*new_args)
        return ChannelList(results)

    @classmethod
    def _check_valid_rate_name(cls, string):
        # Added check, not in sclang.
        if string not in cls._valid_rates:
            raise ValueError(f"{cls.__name__} invalid rate: '{string}'")

    def _init_ugen(self, *inputs):
        '''
        This method is called by _new1 that uses its return value. It must
        return self or ChannelList (cases of MultiOutUGen). Optimizations
        returning scalars or None (for no output) are usually returned by
        public UGen constructors (ar, kr, dr, ir or new).
        '''
        self._inputs = inputs
        return self

    @classmethod
    def _new_from_desc(cls, rate, num_outputs, inputs, special_index):
        obj = cls._create_ugen_object(rate)
        obj._inputs = tuple(inputs)
        obj._special_index = special_index
        return obj

    def __copy__(self):
        # // You can't really copy a UGen without disturbing the Synth.
        # // Usually you want the same object.
        return self


    ### SynthDef build ###

    def _add_to_synth(self):
        self._synthdef = _libsc3.main._current_synthdef
        if self._synthdef is not None:
            self._synthdef._add_ugen(self)

    def _collect_constants(self): # pong
        for input in self.inputs:
            if isinstance(input, (int, float)):
                self._synthdef._add_constant(float(input))

    def _check_inputs(self):  # pong
        '''Returns error msg or None.'''
        return self._check_valid_inputs()

    def _check_valid_inputs(self):
        '''Returns error msg or None.'''
        for i, input in enumerate(self.inputs):
            if not gpp.ugen_param(input)._is_valid_ugen_input():
                arg_name = self._arg_name_for_input_at(i)
                if arg_name is None: arg_name = i
                return f'arg: {arg_name} has bad input: {input}'
        return None

    def _check_n_inputs(self, n):
        if self.rate == 'audio':
            if n > len(self.inputs):
                n = len(self.inputs)
            for i in range(n):
                if gpp.ugen_param(self.inputs[i])._as_ugen_rate() != 'audio':
                    return (f'input {i} is not audio rate: {self.inputs[i]} '
                            f'{gpp.ugen_param(self.inputs[i])._as_ugen_rate()}')
        return self._check_valid_inputs()

    def _check_sr_as_first_input(self):  # Was checkSameRateAsFirstInput.
        if self.rate != gpp.ugen_param(self.inputs[0])._as_ugen_rate():
            return (f'first input is not {self.rate} rate: {self.inputs[0]} '
                    f'{gpp.ugen_param(self.inputs[0])._as_ugen_rate()}')
        return self._check_valid_inputs()

    def _arg_name_for_input_at(self, i):
        try:
            selector = type(self)._method_selector_for_rate(self.rate)
            method = getattr(type(self), selector)
            sig = inspect.signature(method)
            params = list(sig.parameters.values())
            arg_names = [x.name for x in params]
            if len(arg_names) == 0:
                return None
            i += self._arg_names_inputs_offset()
            if i < len(arg_names):
                return arg_names[i]
            else:
                return None
        except AttributeError:
            return None

    def _arg_names_inputs_offset(self):
        '''
        This method exist for two reasons, one is to skip 'this' argument in
        sclang introspection which is not needed in Python, the other reason is
        that many 'spec' ugens, such as EnvGen, Klang, etc., receive the
        specification array as the first argument in the language constructors
        but becomes a variable length list of arguments at the end of the
        server's ugen instruction so is moved at the end of the stored inputs.
        '''
        return 0  # NOTE: Is one less than sclang.

    @classmethod
    def _method_selector_for_rate(cls, rate):
        if rate == 'audio' or rate == 'ar':
            if hasattr(cls, 'ar'):
                return 'ar'
        elif rate == 'control' or rate == 'kr':
            if hasattr(cls, 'kr'):
                return 'kr'
        elif rate == 'scalar' or rate == 'ir':
            if hasattr(cls, 'ir'):
                return 'ir'
        elif rate == 'demand' or rate == 'dr':
            if hasattr(cls, 'dr'):
                return 'dr'
        elif rate is None:
            if hasattr(cls, 'new'):
                return 'new'
        # return None  # original behaviour
        raise AttributeError(f'{cls.__name__} has no {rate} rate constructor')

    def _dump_args(self):
        '''Used for error messages.'''
        msg = 'ARGS:\n'
        tab = ' ' * 4
        arg_name = None
        for i, input in enumerate(self.inputs):
            arg_name = self._arg_name_for_input_at(i)
            if arg_name is None:
                arg_name = str(i)
            msg += tab + arg_name + ': ' + str(input)
            msg += ' ' + type(input).__name__ + '\n'
        print(msg, end='')

    def _dump_name(self):
        '''Used for SynthDef.dump_ugens().'''
        return str(self._synth_index) + '_' + self.name


    @classmethod
    def _replace_zeroes_with_silence(cls, lst):
        # // This replaces zeroes with audio rate silence.
        # // Sub collections are deep replaced.
        num_zeroes = lst.count(0.0)
        if num_zeroes == 0:
            return lst
        silent_channels = ChannelList(lne.Silent.ar(num_zeroes))
        pos = 0
        for i, item in enumerate(lst):
            if item == 0.0:
                lst[i] = silent_channels[pos]
                pos += 1
            elif isinstance(item, list):
                res = cls._replace_zeroes_with_silence(item)
                lst[i] = res
        return lst


    ### SynthDef binary format ###

    def _write_def(self, file):
        try:
            frw.write_pascal_str(file, self.name)
            frw.write_i8(file, self._rate_number())
            frw.write_i32(file, self._num_inputs())
            frw.write_i32(file, self._num_outputs())
            frw.write_i16(file, self._special_index)
            # // write wire spec indices.
            for input in self.inputs:
                gpp.ugen_param(input)._write_input_spec(file, self._synthdef)
            self._write_output_specs(file)
        except Exception as e:
            raise Exception('SynthDef: could not write def') from e

    @property
    def name(self):  # Was a method, see OutputPorxy
        return type(self).__name__

    def _rate_number(self):
        if self.rate == 'audio': return 2
        if self.rate == 'control': return 1
        if self.rate == 'demand': return 3
        return 0  # 'scalar'

    def _num_inputs(self):
        return len(self.inputs)

    def _num_outputs(self):
        return 1

    def _write_input_spec(self, file, synthdef):
        frw.write_i32(file, self._synth_index)
        frw.write_i32(file, self._output_index)

    def _write_output_spec(self, file):
        frw.write_i8(file, self._rate_number())

    def _write_output_specs(self, file): # TODO: variación con 's' que llama a la sin 's', este método sería para las ugens con salidas múltiples, el nombre del método debería ser más descriptivo porque es fácil de confundir, además. # lo implementan AbstractOut, MultiOutUGen, SendPeakRMS, SendTrig y UGen.
        self._write_output_spec(file)


    ### Topo sort methods ###

    def _init_topo_sort(self):  # pong
        for input in self.inputs:
            if isinstance(input, UGen):
                if isinstance(input, OutputProxy):
                    ugen = input.source_ugen
                else:
                    ugen = input
                self._antecedents.add(ugen)
                ugen._descendants.add(self)
        for ugen in self._width_first_antecedents:
            self._antecedents.add(ugen)
            ugen._descendants.add(self)

    def _make_available(self):
        if len(self._antecedents) == 0:
            self._synthdef._available.append(self)

    def _remove_antecedent(self, ugen):
        self._antecedents.remove(ugen)
        self._make_available()

    def _arrange(self, out_stack):  # Was schedule.
        descendants = list(self._descendants)
        descendants.sort(key=lambda x: x._synth_index)
        for ugen in reversed(descendants):
            ugen._remove_antecedent(self)
        out_stack.append(self)

    def _optimize_graph(self):  # pong
        pass

    def _perform_dead_code_elimination(self):
        if len(self._descendants) == 0:
            # for input in self.inputs:  # It should be _antecedents, check sclang.
            for input in self._antecedents:
                if isinstance(input, UGen):
                    input._descendants.remove(self)
                    input._optimize_graph()
            self._synthdef._remove_ugen(self)
            return True
        return False


    ### SynthDesc interface ###

    # def _writes_to_bus(self):
    #     return False

    # def _can_free_synth(self):  # Non core interface, removed.
    #     return False

    # def is_control_ugen(cls):  # Use issubclass(cls, AbstractControl).
    # def is_input_ugen(cls):  # Use issubclass(cls, AbstractIn).
    # def is_output_ugen(cls):  # Use issubclass(cls, AbstractOut).
    # def is_ugen(self):  # Use isinstance(obj, UGen) (not used in sclang).
    # def _output_index(self): # Is attribute now.


    ### UGen graph parameter interface ###

    def _is_valid_ugen_input(self):
        return True

    def _as_ugen_input(self, *ugen_cls):
        return self

    # def _as_control_input(self):  # Is NodeParameter interface.
    #     raise TypeError("UGen can't be set as control input")

    def _as_audio_rate_input(self):
        if self.rate != 'audio':
            return lne.K2A.ar(self)
        return self

    def _as_ugen_rate(self): # Was rate.
        return self.rate


class UGen(SynthObject, aob.AbstractObject):
    @classmethod
    def signal_range(cls):
        return 'bipolar'


    ### AbstractObject interface ###

    def _compose_unop(self, selector):
        selector = _si.sc_opname(selector.__name__)
        return UnaryOpUGen.new(selector, self)

    def _compose_binop(self, selector, input):
        param = gpp.ugen_param(input)
        if param._is_valid_ugen_input():
            selector = _si.sc_opname(selector.__name__)
            return BinaryOpUGen.new(selector, self, input)
        else:
            return param._perform_binary_op_on_ugen(selector, self)

    def _rcompose_binop(self, selector, input):
        param = gpp.ugen_param(input)
        if param._is_valid_ugen_input():
            selector = _si.sc_opname(selector.__name__)
            return BinaryOpUGen.new(selector, input, self)
        else:
            return param._r_perform_binary_op_on_ugen(selector, self)

    def _compose_narop(self, selector, *args):
        raise NotImplementedError('UGen _compose_narop is not supported')

    # L426
    # // Complex support
    # asComplex
    # performBinaryOpOnComplex
    # def _perform_binary_op_on_ugen(input, selector, thing):


    ### Convenience methods (sync with ChannelList) ###

    def dup(self, n=2):
        return ChannelList([self] * n)

    def madd(self, mul=1.0, add=0.0):
        return MulAdd.new(self, mul, add)

    def range(self, lo=0.0, hi=1.0):
        if type(self).signal_range() == 'bipolar':
            mul = (hi - lo) * 0.5
            add = mul + lo
        else:
            mul = (hi - lo)
            add = lo
        return MulAdd.new(self, mul, add)

    def exprange(self, lo=0.01, hi=1.0):
        if type(self).signal_range() == 'bipolar':
            return self.linexp(-1, 1, lo, hi, None)
        else:
            return self.linexp(0, 1, lo, hi, None)

    def curverange(self, lo=0.0, hi=1.0, curve=-4):
        if type(self).signal_range() == 'bipolar':
            return self.lincurve(-1, 1, lo, hi, curve, None)
        else:
            return self.lincurve(0, 1, lo, hi, curve, None)

    def unipolar(self, mul=1):
        return self.range(0, mul)

    def bipolar(self, mul=1):
        return self.range(-mul, mul)

    def clip(self, lo=0.0, hi=1.0):
        if self.rate == 'demand':
            bi.max(lo, bi.min(hi, self))
        else:
            selector = trg.Clip._method_selector_for_rate(self.rate)
            return getattr(trg.Clip, selector)(self, lo, hi)

    def fold(self, lo=0.0, hi=0.0):
        if self.rate == 'demand':
            raise NotImplementedError('fold is not implemented for dr ugens')
        else:
            selector = trg.Fold._method_selector_for_rate(self.rate)
            return getattr(trg.Fold, selector)(self, lo, hi)

    def wrap(self, lo=0.0, hi=1.0):
        if self.rate == 'demand':
            raise NotImplementedError('wrap is not implemented for dr ugens')
        else:
            selector = trg.Wrap._method_selector_for_rate(self.rate)
            return getattr(trg.Wrap, selector)(self, lo, hi)

    def degrad(self):  # override (not to call bi.degrad)
        return self * (bi.pi / 180.)

    def raddeg(self):  # override (not to call bi.raddeg)
        return self * (180. / bi.pi)

    def blend(self, other, frac=0.5):
        if self.range == 'demand' or gpp.ugen_param(other).rate == 'demand':
            raise NotImplementedError('blend is not implemented for dr ugens')
        else:
            pan = bi.linlin(frec, 0.0, 1.0, -1.0, 1.0)
            if self.rate == 'audio':
                return pan.XFade2.ar(self, other, pan)
            if gpp.ugen_param(other).rate == 'audio':
                return pan.XFade2.ar(other, self, -pan)
            selector = pan.LinXFade2._method_selector_for_rate(self.rate)
            return getattr(pan.LinXFade2, selector)(self, other, pan)

    def min_nyquist(self):
        return bi.min(self, ifu.SampleRate.ir * 0.5)

    def lag(self, time=0.1):
        selector = flr.Lag._method_selector_for_rate(self.rate)
        return getattr(flr.Lag, selector)(self, time)

    def lag2(self, time=0.1):
        selector = flr.Lag2._method_selector_for_rate(self.rate)
        return getattr(flr.Lag2, selector)(self, time)

    def lag3(self, time=0.1):
        selector = flr.Lag3._method_selector_for_rate(self.rate)
        return getattr(flr.Lag3, selector)(self, time)

    def lagud(self, utime=0.1, dtime=0.1):
        selector = flr.LagUD._method_selector_for_rate(self.rate)
        return getattr(flr.LagUD, selector)(self, utime, dtime)

    def lag2ud(self, utime=0.1, dtime=0.1):
        selector = flr.Lag2UD._method_selector_for_rate(self.rate)
        return getattr(flr.Lag2UD, selector)(self, utime, dtime)

    def lag3ud(self, utime=0.1, dtime=0.1):
        selector = flr.Lag3UD._method_selector_for_rate(self.rate)
        return getattr(flr.Lag3UD, selector)(self, utime, dtime)

    def varlag(self, time=0.1, curvature=0, wrap=5, start=None):
        selector = flr.VarLag._method_selector_for_rate(self.rate)
        return getattr(flr.VarLag, selector)(self, time, curvature, wrap, start)

    def slew(self, up=1, down=1):
        selector = flr.Slew._method_selector_for_rate(self.rate)
        return getattr(flr.Slew, selector)(self, up, down)

    def prune(self, min, max, type='minmax'):
        if type == 'minmax':
            return self.clip(min, max)
        elif type == 'min':
            return self.max(min)
        elif type == 'max':
            return self.min(max)
        return self

    def snap(self, resolution=1.0, margin=0.05, strengh=1.0):  # NOTE: UGen/SimpleNumber, not in AbstractFunction
        selector = ocl.Select._method_selector_for_rate(self.rate)
        diff = round(self, resolution) - self
        return getattr(ocl.Select, selector)(
            abs(diff) < margin, [self, self + strengh * diff])

    def softround(self, resolution=1.0, margin=0.05, strengh=1.0):  # NOTE: UGen/SimpleNumber, not in AbstractFunction
        selector = ocl.Select._method_selector_for_rate(self.rate)
        diff = round(self, resolution) - self
        return getattr(ocl.Select, selector)(
            abs(diff) > margin, [self, self + strengh * diff])

    def linlin(self, inmin, inmax, outmin, outmax, clip='minmax'):
        selector = lne.LinLin._method_selector_for_rate(self.rate)  # BUG: I see these can fail for ir/dr ugens however sclang implementation semantics is diverse and not clear.
        return getattr(lne.LinLin, selector)(
            self.prune(inmin, inmax, clip), inmin, inmax, outmin, outmax)

    def linexp(self, inmin, inmax, outmin, outmax, clip='minmax'):
        selector = lne.LinExp._method_selector_for_rate(self.rate)
        return getattr(lne.LinExp, selector)(
            self.prune(inmin, inmax, clip), inmin, inmax, outmin, outmax)

    def explin(self, inmin, inmax, outmin, outmax, clip='minmax'):
        return (
            bi.log(self.prune(inmin, inmax, clip) / inmin) /
            bi.log(inmax / inmin) * (outmax - outmin) + outmin)  # // no separate ugen yet

    def expexp(self, inmin, inmax, outmin, outmax, clip='minmax'):
        return pow(
            outmax / outmin,
            bi.log(self.prune(inmin, inmax, clip) / inmin) /
            bi.log(inmax / inmin)) * outmin

    def lincurve(self, inmin, inmax, outmin, outmax, curve=-4, clip='minmax'):
        if isinstance(curve, (int, float)) and abs(curve) < 0.125:
            return self.linlin(inmin, inmax, outmin, outmax, clip)
        grow = bi.exp(curve)
        a = (outmax - outmin) / (1.0 - grow)
        b = outmin + a
        scaled = (self.prune(inmin, inmax, clip) - inmin) / (inmax - inmin)
        curved_res = b - a * pow(grow, scaled)
        if gpp.ugen_param(curve).rate == 'scalar':
            return curved_res
        else:
            selector = ocl.Select._method_selector_for_rate(self.rate)
            return getattr(ocl.Select, selector)(abs(curve) >= 0.125,
                [self.linlin(inmin, inmax, outmin, outmax, clip), curved_res])

    def curvelin(self, inmin, inmax, outmin, outmax, curve=-4, clip='minmax'):
        if isinstance(curve, (int, float)) and abs(curve) < 0.125:
            return self.linlin(inmin, inmax, outmin, outmax, clip)
        grow = bi.exp(curve)
        a = (inmax - inmin) / (1.0 - grow)
        b = inmin + a
        lin_res = (
            bi.log((b - this.prune(inmin, inmax, clip)) / a) *
            (outmax - outmin) / curve + outmin)
        if gpp.ugen_param(curve).rate == 'scalar':
            return lin_res
        else:
            selector = ocl.Select._method_selector_for_rate(self.rate)
            return getattr(ocl.Select, selector)(abs(curve) >= 0.125,
                [self.linlin(inmin, inmax, outmin, outmax, clip), lin_res])

    def bilin(self, incenter, inmin, inmax, outcenter, outmin, outmax,
              clip='minmax'):
        selector = ocl.Select._method_selector_for_rate(self.rate)  # BUG: in sclang the call is over the wrong class and doesn't uses _multi_new as above.
        return getattr(ocl.Select, selector)(self < incenter, [
            self.linlin(incenter, inmax, outcenter, outmax, clip),
            self.linlin(inmin, incenter, outmin, outcenter, clip)])

    # biexp is not overridden

    def moddif(self, that=0.0, mod=1.0):
        selector = trg.ModDif._method_selector_for_rate(self.rate)
        return getattr(trg.ModDif, selector)(self, that, mod)

    def sanitize(self):
        selector = tsu.Sanitize._method_selector_for_rate(self.rate)  # BUG: in sclang the call is over the wrong class.
        return getattr(tsu.Sanitize, selector)(self)

    # degreeToKey (I don't know why this method is important)

    # Synth debug convenience methods

    def poll(self, trig=10, label=None, trig_id=-1):
        return pll.Poll.new(trig, self, label, trig_id)

    def dpoll(self, label=None, run=1, trig_id=-1):
        return dmd.Dpoll.new(self, label, run, trig_id)

    def check_bad_values(self, id=0, post=2):
        selector = tsu.CheckBadValues._method_selector_for_rate(self.rate)
        getattr(tsu.CheckBadValues, selector)(self, id, post)
        # // add the UGen to the tree but keep self as the output
        return self


class PureUGen():
    # // UGen which has no side effect and can therefore be considered for
    # // a dead code elimination. Read access to buffers/busses are allowed.
    def _optimize_graph(self):  # override
        self._perform_dead_code_elimination()


class MultiOutUGen(UGen):
    @property
    def _synth_index(self):
        return self.__synth_index

    @_synth_index.setter
    def _synth_index(self, value):
        self.__synth_index = value
        for output in self._channels:
            output._synth_index = value

    @_synth_index.deleter
    def _synth_index(self):
        del self.__synth_index

    @classmethod
    def _new_from_desc(cls, rate, num_outputs, inputs, special_index=None):  # override
        obj = cls._create_ugen_object(rate)
        obj._inputs = tuple(inputs)
        obj._init_outputs(num_outputs, rate)
        return obj

    def _init_outputs(self, num_channels, rate):
        '''
        Return value of this method is used as return value of _init_ugen
        in subclasses.
        '''
        if num_channels is None or num_channels < 1:
            raise Exception(
                f'{self.name}: wrong number of channels ({num_channels})')
        self._channels = ChannelList(
            [OutputProxy.new(rate, self, i) for i in range(num_channels)])
        if num_channels == 1:
            return self._channels[0]
        return self._channels

    def _num_outputs(self):  # override
        return len(self._channels)

    def _write_output_specs(self, file):  # override
        for output in self._channels:
            output._write_output_spec(file)


class OutputProxy(UGen):
    @classmethod
    def new(cls, rate, source_ugen, index):
        return cls._new1(rate, source_ugen, index)

    def _init_ugen(self, source_ugen, index):  # override
        self.source_ugen = source_ugen  # Was just source.
        self._output_index = index
        self._synth_index = source_ugen._synth_index
        return self  # NOTE: Must return self.

    def _add_to_synth(self):  # override
        # OutputProxy is not part of the SynthDef graph but source_ugen.
        self._synthdef = _libsc3.main._current_synthdef

    def _dump_name(self):  # override
        return f'{self.source_ugen._dump_name()}[{self._output_index}]'

    @property
    def name(self):
        # NOTE: OutputProxy define <>name, Control UGen return OutputPorxy
        # in SynthDef _set_control_names and change this property but can't
        # find where this getter is used (if used).
        try:
            return self.__name
        except AttributeError:
            return None

    @name.setter
    def name(self, value):
        self.__name = value


class WidthFirstUGen(SynthObject):  # Was in fft.py
    _default_rate = None
    # bufio.py uses new to create 'scalar'
    # fft uses new to create 'control'

    def _add_to_synth(self):  # override
        self._synthdef = _libsc3.main._current_synthdef
        if self._synthdef is not None:
            self._synthdef._add_ugen(self)
            self._synthdef._width_first_ugens.append(self)

    def _add_copies_if_needed(self):
        pass


### BasicOpUGens.sc ###

class BasicOpUGen(UGen):
    def __init__(self):
        super().__init__()
        self._operator = None

    # writeName commented method, no other standard UGen class defines it.

    @property
    def operator(self):
        return self._operator

    @operator.setter
    def operator(self, value):
        index, operator = _si.sc_spindex_opname(value)
        self._operator = operator
        self._special_index = index
        if self._special_index < 0:
            raise Exception(
                f"operator '{value}' applied to a UGen "
                "is not supported by the server")

    @operator.deleter
    def operator(self):
        del self._operator

    def _arg_name_for_input_at(self, i):  # override
        try:
            method = getattr(type(self), 'new')
            sig = inspect.signature(method)
            params = list(sig.parameters.values())
            arg_names = [x.name for x in params]
            if len(arg_names) == 0:
                return None
            i += self._arg_names_inputs_offset()
            if i < len(arg_names):
                return arg_names[i]
            else:
                return None
        except AttributeError:
            return None

    def _arg_names_inputs_offset(self):  # override
        return 1  # One less than sclang

    def _dump_args(self):  # override
        msg = 'ARGS:\n'
        tab = ' ' * 4
        msg += tab + 'operator: ' + self.operator + '\n'
        arg_name = None
        for i, input in enumerate(self.inputs):
            arg_name = self._arg_name_for_input_at(i)
            if not arg_name:
                arg_name = str(i)
            msg += tab + arg_name + ': ' + str(input)
            msg += ' ' + type(input).__name__ + '\n'
        print(msg, end='')

    def _dump_name(self):  # override
        return str(self._synth_index) + '_' + self.operator


class UnaryOpUGen(BasicOpUGen):
    @classmethod
    def new(cls, selector, a):
        return cls._multi_new('audio', selector, a)

    @classmethod
    def _new_from_desc(cls, rate, num_outputs, inputs, special_index):  # override
        # *** BUG: this method is missing in sclang
        obj = super()._new_from_desc(rate, num_outputs, inputs, special_index)
        obj._operator = _si.sc_opname_from_index(special_index, 'unary')
        return obj

    def _init_ugen(self, operator, input):  # override
        self.operator = operator
        self._rate = gpp.ugen_param(input)._as_ugen_rate()
        self._inputs = (input,)
        return self  # Must return self.

    def _optimize_graph(self):  # override
        self._perform_dead_code_elimination()


class BinaryOpUGen(BasicOpUGen):
    @classmethod
    def _new1(cls, rate, selector, a, b):  # override
        if selector == '*':
            if a == 0.0: return 0.0
            if b == 0.0: return 0.0
            if a == 1.0: return b
            if a == -1.0: return -b  # neg
            if b == 1.0: return a
            if b == -1.0: return -a  # neg
        if selector == '+':
            if a == 0.0: return b
            if b == 0.0: return a
        if selector == '-':
            if a == 0.0: return -b  # neg
            if b == 0.0: return a
        if selector == '/':
            if b == 1.0: return a
            if b == -1.0: return -a  # neg
        return super()._new1(rate, selector, a, b)

    @classmethod
    def new(cls, selector, a, b):
        return cls._multi_new('audio', selector, a, b)

    @classmethod
    def _new_from_desc(cls, rate, num_outputs, inputs, special_index):  # override
        obj = super()._new_from_desc(rate, num_outputs, inputs, special_index)
        obj._operator = _si.sc_opname_from_index(special_index, 'binary')
        return obj

    def _init_ugen(self, operator, a, b):  # override
        self.operator = operator
        self._rate = self._determine_rate(a, b)
        self._inputs = (a, b)
        return self  # Must return self.

    def _determine_rate(self, a, b):
        a_rate = gpp.ugen_param(a)._as_ugen_rate()
        b_rate = gpp.ugen_param(b)._as_ugen_rate()
        # Order matters.
        if a_rate == 'demand': return 'demand'
        if b_rate == 'demand': return 'demand'
        if a_rate == 'audio': return 'audio'
        if b_rate == 'audio': return 'audio'
        if a_rate == 'control': return 'control'
        if b_rate == 'control': return 'control'
        return 'scalar'

    def _optimize_graph(self):  # override
        # // this.constantFolding;
        if self._perform_dead_code_elimination():
            return self
        if self.operator == '+':
            self._optimize_add()
            return self
        if self.operator == '-':
            self._optimize_sub()
            return self

    def _optimize_add(self):
        # // create a Sum3 if possible
        optimized_ugen = self._optimize_to_sum3()
        # // create a Sum4 if possible
        if not optimized_ugen:
            optimized_ugen = self._optimize_to_sum4()
        # // create a MulAdd if possible.
        if not optimized_ugen:
            optimized_ugen = self._optimize_to_muladd()
        # // optimize negative additions
        if not optimized_ugen:
            optimized_ugen = self._optimize_addneg()

        if optimized_ugen:
            self._synthdef._replace_ugen(self, optimized_ugen)

    def _optimize_to_sum3(self):
        a, b = self.inputs
        if gpp.ugen_param(a)._as_ugen_rate() == 'demand'\
        or gpp.ugen_param(b)._as_ugen_rate() == 'demand':
            return None

        if isinstance(a, BinaryOpUGen) and a.operator == '+'\
        and len(a._descendants) == 1:
            self._synthdef._remove_ugen(a)
            if a is b:  # Edge case fixed in supercollider/supercollider#5048
                replacement = Sum4.new(
                    a.inputs[0], a.inputs[0], a.inputs[1], a.inputs[1])
            else:
                replacement = Sum3.new(a.inputs[0], a.inputs[1], b)
            replacement._descendants = self._descendants
            self._optimize_update_descendants(replacement, a)
            return replacement

        if isinstance(b, BinaryOpUGen) and b.operator == '+'\
        and len(b._descendants) == 1:
            self._synthdef._remove_ugen(b)
            replacement = Sum3.new(b.inputs[0], b.inputs[1], a)
            replacement._descendants = self._descendants
            self._optimize_update_descendants(replacement, b)
            return replacement

        return None

    def _optimize_to_sum4(self):
        a, b = self.inputs
        if gpp.ugen_param(a)._as_ugen_rate() == 'demand'\
        or gpp.ugen_param(b)._as_ugen_rate() == 'demand':
            return None

        if isinstance(a, Sum3) and len(a._descendants) == 1:
            self._synthdef._remove_ugen(a)
            replacement = Sum4.new(a.inputs[0], a.inputs[1], a.inputs[2], b)
            replacement._descendants = self._descendants
            self._optimize_update_descendants(replacement, a)
            return replacement

        if isinstance(b, Sum3) and len(b._descendants) == 1:
            self._synthdef._remove_ugen(b)
            replacement = Sum4.new(b.inputs[0], b.inputs[1], b.inputs[2], a)
            replacement._descendants = self._descendants
            self._optimize_update_descendants(replacement, b)
            return replacement

        return None

    def _optimize_to_muladd(self):
        a, b = self.inputs

        if isinstance(a, BinaryOpUGen) and a.operator == '*'\
        and len(a._descendants) == 1:

            if MulAdd._can_be_muladd(a.inputs[0], a.inputs[1], b):
                self._synthdef._remove_ugen(a)
                replacement = MulAdd.new(a.inputs[0], a.inputs[1], b)
                replacement._descendants = self._descendants
                self._optimize_update_descendants(replacement, a)
                return replacement

            if MulAdd._can_be_muladd(a.inputs[1], a.inputs[0], b):
                self._synthdef._remove_ugen(a)
                replacement = MulAdd.new(a.inputs[1], a.inputs[0], b)
                replacement._descendants = self._descendants
                self._optimize_update_descendants(replacement, a)
                return replacement

        # does optimization code need to be optimized?
        if isinstance(b, BinaryOpUGen) and b.operator == '*'\
        and len(b._descendants) == 1:

            if MulAdd._can_be_muladd(b.inputs[0], b.inputs[1], a):
                self._synthdef._remove_ugen(b)
                replacement = MulAdd.new(b.inputs[0], b.inputs[1], a)
                replacement._descendants = self._descendants
                self._optimize_update_descendants(replacement, b)
                return replacement

            if MulAdd._can_be_muladd(b.inputs[1], b.inputs[0], a):
                self._synthdef._remove_ugen(b)
                replacement = MulAdd.new(b.inputs[1], b.inputs[0], a)
                replacement._descendants = self._descendants
                self._optimize_update_descendants(replacement, b)
                return replacement

        return None

    def _optimize_addneg(self):
        a, b = self.inputs

        if isinstance(b, UnaryOpUGen) and b.operator == 'neg'\
        and len(b._descendants) == 1:
            # // a + b.neg -> a - b
            self._synthdef._remove_ugen(b)
            replacement = a - b.inputs[0]
            # // This is the first time the dependants logic appears. It's
            # // repeated below. We will remove 'self' from the synthdef, and
            # // replace it with 'replacement'. 'replacement' should then have
            # // all the same descendants as 'self'.
            replacement._descendants = self._descendants
            # // Drop 'self' and 'b' from all of replacement's inputs'
            # // descendant lists so that future optimizations decide correctly.
            self._optimize_update_descendants(replacement, b)
            return replacement

        if isinstance(a, UnaryOpUGen) and a.operator == 'neg'\
        and len(a._descendants) == 1:
            # // a.neg + b -> b - a
            self._synthdef._remove_ugen(a)
            replacement = b - a.inputs[0]
            replacement._descendants = self._descendants
            self._optimize_update_descendants(replacement, a)
            return replacement

        return None

    def _optimize_sub(self):
        a, b = self.inputs

        if isinstance(b, UnaryOpUGen) and b.operator == 'neg'\
        and len(b._descendants) == 1:
            # // a - b.neg -> a + b
            self._synthdef._remove_ugen(b)
            replacement = BinaryOpUGen.new('+', a, b.inputs[0])
            replacement._descendants = self._descendants
            self._optimize_update_descendants(replacement, b)
            self._synthdef._replace_ugen(self, replacement)
            replacement._optimize_graph()  # // Not called from _optimize_add, no need to return ugen here.

        return None

    def _optimize_update_descendants(self, replacement, deleted_unit):
        # // 'this' = old ugen being replaced
        # // replacement = this's replacement
        # // deletedUnit = auxiliary unit being removed, not replaced
        for input in replacement.inputs:
            if isinstance(input, UGen):
                if isinstance(input, OutputProxy):
                    input = input.source_ugen
                if input._descendants is None:
                    return
                input._descendants.add(replacement)
                input._descendants.discard(self)
                input._descendants.discard(deleted_unit)

    def _constant_folding(self): # Not sure if used.
        ... # BUG, boring to copy.


class MulAdd(UGen):
    @classmethod
    def new(cls, input, mul=1.0, add=0.0):
        params = gpp.ugen_param([input, mul, add])
        rate = params._as_ugen_rate()
        args = params._as_ugen_input(cls)
        return cls._multi_new_list([rate] + args)

    @classmethod
    def _new1(cls, rate, input, mul, add):  # override
        if mul == 0.0: return add
        minus = mul == -1.0
        nomul = mul == 1.0
        noadd = add == 0.0
        if nomul and noadd: return input
        if minus and noadd: return -input  # neg
        if noadd: return input * mul
        if minus: return add - input
        if nomul: return input + add

        if cls._can_be_muladd(input, mul, add):
            return super()._new1(rate, input, mul, add)
        if cls._can_be_muladd(mul, input, add):
            return super()._new1(rate, mul, input, add)
        return (input * mul) + add

    def _init_ugen(self, input, mul, add):  # override
        self._inputs = (input, mul, add)
        self._rate = gpp.ugen_param(self.inputs)._as_ugen_rate()
        return self  # Must return self.

    @classmethod
    def _can_be_muladd(cls, input, mul, add):
        # // see if these inputs satisfy the constraints of a MulAdd ugen.
        in_rate = gpp.ugen_param(input)._as_ugen_rate()
        if in_rate == 'audio':
            return True
        mul_rate = gpp.ugen_param(mul)._as_ugen_rate()
        add_rate = gpp.ugen_param(add)._as_ugen_rate()
        if in_rate == 'control'\
        and (mul_rate == 'control' or mul_rate == 'scalar')\
        and (add_rate == 'control' or add_rate == 'scalar'):
            return True
        return False


class Sum3(UGen):
    @classmethod
    def new(cls, in0, in1, in2):
        return cls._multi_new(None, in0, in1, in2)

    @classmethod
    def _new1(cls, _, in0, in1, in2):  # override
        if in2 == 0.0: return in0 + in1
        if in1 == 0.0: return in0 + in2
        if in0 == 0.0: return in1 + in2

        arg_list = [in0, in1, in2]
        rate = gpp.ugen_param(arg_list)._as_ugen_rate()
        arg_list.sort(key=lambda x: gpp.ugen_param(x)._as_ugen_rate())  # NOTE: Why sort?

        return super()._new1(rate, *arg_list)


class Sum4(UGen):
    @classmethod
    def new(cls, in0, in1, in2, in3):
        return cls._multi_new(None, in0, in1, in2, in3)

    @classmethod
    def _new1(cls, _, in0, in1, in2, in3):  # override
        if in0 == 0.0: return Sum3._new1(None, in1, in2, in3)
        if in1 == 0.0: return Sum3._new1(None, in0, in2, in3)
        if in2 == 0.0: return Sum3._new1(None, in0, in1, in3)
        if in3 == 0.0: return Sum3._new1(None, in0, in1, in2)

        arg_list = [in0, in1, in2, in3]
        rate = gpp.ugen_param(arg_list)._as_ugen_rate()
        arg_list.sort(key=lambda x: gpp.ugen_param(x)._as_ugen_rate())  # NOTE: Why sort?

        return super()._new1(rate, *arg_list)


class PseudoUGen(SynthObject):
    # Base class to reinforce the interface. Pseudo UGens never
    # return instances of themselves but of another UGen subclass.
    pass



"""SynthDef.sc"""

import inspect
import logging
import io
import pathlib
from collections.abc import Container

from ..base import classlibrary as clb
from ..base import utils as utl
from ..base import platform as plf
from ..base import systemactions as sac
from ..base import functions as fn
from ..base import main as _libsc3
from . import spec as spc
from . import ugen as ugn
from . import server as srv
from . import synthdesc as sdc
from . import _fmtrw as frw
from . import _graphparam as gpp
from . import node as nod
from .ugens import inout as iou
from .ugens import fftunpacking as ffu


__all__ = ['SynthDef', 'synthdef']


_logger = logging.getLogger(__name__)


class MetaSynthDef(type):
    tmp_name_prefix = 'tmp__'
    _tmp_def_count = 0

    def __init__(cls, *_):

        def init_func(cls):
            # // Ensure exists.
            if not plf.Platform.support_dir.exists():
                _logger.info(
                    'creating Platform.support_dir '
                    f'at {plf.Platform.support_dir}')
            plf.Platform.synthdef_dir.mkdir(parents=True, exist_ok=True)

        clb.ClassLibrary.add(cls, init_func)

    def generate_tmp_name(cls):
        name = cls.tmp_name_prefix + str(cls._tmp_def_count)
        cls._tmp_def_count += 1
        return name


class SynthDef(metaclass=MetaSynthDef):
    """Build the instructions to create synthesis nodes in the server.

    SynthDef instances build the instructions to create synthesis nodes
    in the server from a function containing interconnected UGen objects.
    In order for the server to have the synthesis definition available it
    should be sent using ``add``, ``load``, ``send`` or ``store`` methods.

    Parameters
    ----------
    name : str
        The name of the synthesis definition used by the server
        to create synth nodes.
    func: function
        A common function containing UGen objects.
    rates : list
        An optional list of rate specifications that map to
        the ``func`` defined parameters. If value is `None` (default)
        control rate instances will be created instead. Possible
        values are `'ar'`, `'kr'`, `'ir'`, `'tr'` or a number
        (indicating lag value for `'kr'` controls). This parameter
        overrides rates defined by function annotations.
    prepend : list
        An optional list of positional values that will be
        passed to ``func`` when evaluated. This prevents controls
        from being created for those parameters.
    variants : dict
        An optional dictionary with different keys that specify
        dictionaries of default values to create synthesis nodes
        in the server. When using variants, synthesis definition
        names are composed as `'synthname.variantkey'`.
    metadata : dict
        An user defined JSON serializable dictionary to
        provide information about the synthesis definition.
    """

    _SUFFIX = 'scsyndef'
    _RATE_NAMES = ('ar', 'kr', 'ir', 'tr')

    @classmethod
    def _dummy(cls, name):
        # Creates an empty object used by SynthDesc.
        obj = cls.__new__(cls)

        obj._name = name
        obj._func = None
        obj._variants = dict()
        obj.metadata = dict()
        obj.desc = None
        obj._bytes = None

        obj._controls = None
        obj._control_names = []
        obj._all_control_names = []
        obj._control_index = 0
        obj._children = []

        obj._constants = dict()
        obj._constant_set = set()
        obj._max_local_bufs = None

        obj._available = []
        obj._width_first_ugens = []
        obj._rewrite_in_progress = False

        return obj

    def __init__(self, name, func, rates=None, prepend=None,
                 variants=None, metadata=None):
        self._name = name
        self._func = None
        self._variants = variants or dict()
        self.metadata = metadata or dict()
        self.desc = None
        self._bytes = None

        # self._controls = None  # init_build, is set by ugens using _libsc3.main._current_synthdef
        self._control_names = []
        self._all_control_names = []
        self._control_index = 0
        self._children = []

        # self._constants = dict()  # init_build
        # self._constant_set = set()  # init_build
        # self._max_local_bufs = None  # init_build, used by LocalBus*new1 check for nil.

        # topo sort
        self._available = []
        self._width_first_ugens = []
        self._rewrite_in_progress = False

        # callable interface
        # self._callable_args = None

        self._build(func, rates or [], prepend or [])

    def _build(self, func, rates, prepend):
        with _libsc3.main._def_build_lock:
            try:
                _libsc3.main._current_synthdef = self
                self._init_build()
                self._build_ugen_graph(func, rates, prepend)
                self._finish_build()
                self._func = func
                _libsc3.main._current_synthdef = None
            except Exception:
                _libsc3.main._current_synthdef = None
                raise

    @property
    def name(self):
        return self._name

    @property
    def func(self):
        return self._func

    @property
    def variants(self):
        return self._variants

    @classmethod
    def wrap(cls, func, rates=None, prepend=None):
        if _libsc3.main._current_synthdef is not None:
            return _libsc3.main._current_synthdef._build_ugen_graph(
                func, rates or [], prepend or [])
        else:
            raise Exception(
                'SynthDef wrap should be called inside '
                'a SynthDef graph function')

    def _init_build(self):
        # UGen.buildSynthDef, lock above.
        self._constants = dict()
        self._constant_set = set()
        self._controls = []
        self._control_index = 0
        self._max_local_bufs = None

    def _build_ugen_graph(self, func, rates, prepend):
        # // Save/restore controls in case of SynthDef.wrap.
        save_ctl_names = self._control_names
        self._control_names = []
        prepend = utl.as_list(prepend)
        self._args_to_controls(func, rates, len(prepend))
        result = func(*(prepend + self._build_controls()))
        self._control_names = save_ctl_names
        return result

    def _args_to_controls(self, func, rates, skip_args=0):  # Was addControlsFromArgsOfFunc.
        if not inspect.isfunction(func):
            raise TypeError('func argument is not a function')

        sig = inspect.signature(func)
        self._callable_args = list(sig.parameters.keys())
        params = list(sig.parameters.values())

        if not params:
            return

        pork = inspect.Parameter.POSITIONAL_OR_KEYWORD
        for p in params:
            if p.kind != pork:
                raise ValueError(
                    'all func parameters must be POSITIONAL_OR_KEYWORD')

        for p in params[skip_args:]:
            if isinstance(p.default, tuple)\
            and any(isinstance(v, Container) for v in p.default):
                raise ValueError(f"tuple rank > 1 for parameter '{p.name}'")

        # // What we do here is separate the ir, tr and kr rate arguments,
        # // create one Control ugen for all of each rate, and then construct
        # // the argument array from combining the OutputProxies of these two
        # // Control ugens in the original order.
        names = [x.name for x in params]
        names = names[skip_args:]
        values = self._get_valid_arg_values(params)
        values = values[skip_args:]
        values = self._apply_metadata_specs(names, values)

        empty = inspect.Signature.empty
        annotations = [
            x.annotation if x.annotation != empty else None for x in params]
        annotations = annotations[skip_args:]

        rates += [0] * (len(names) - len(rates))
        rates = [x if x is not None else 0.0 for x in rates]

        rate_names = self._RATE_NAMES

        for a in annotations:
            if a is not None and a not in rate_names:
                raise ValueError(
                    f'rate annotation {repr(a)} not in {rate_names}')

        for i, name in enumerate(names):
            value = values[i]
            annot = annotations[i]
            lag = rates[i]

            if annot and annot != 'kr'\
            and isinstance(lag, (int, float)) and lag != 0:
                _logger.warning(
                    f"lag value {lag} for '{annot}' parameter "
                    f"'{name}' will be ignored")

            overridden = lag in rate_names

            if lag == 'ir' or annot == 'ir' and not overridden:
                self._add_ir(name, value)
            elif lag == 'tr' or annot == 'tr' and not overridden:
                self._add_tr(name, value)
            elif lag == 'ar' or annot == 'ar' and not overridden:
                self._add_ar(name, value)
            else:
                if lag == 'kr': lag = 0.0
                self._add_kr(name, value, lag)

    def _get_valid_arg_values(self, params):
        valid_args = (int, float, bool, complex, type(None))  # + tuple
        ret = []
        empty = inspect.Signature.empty
        for param in params:
            if param.default != empty:
                if isinstance(param.default, valid_args):
                    ret.append(param.default)
                elif isinstance(param.default, tuple):
                    ret.append(list(param.default))
                else:
                    _logger.warning(
                        "invalid value as default argument, "
                        f"'{param.default}' replaced by None")
                    ret.append(None)
            else:
                ret.append(None)
        return ret

    def _apply_metadata_specs(self, names, values):
        # This method depends on MdPlugin 'specs' key.
        # I'm not sure if SynthDef is the right place for it
        # but it's called when building the definition and may add
        # a default value for control parameters if they are None.
        new_values = []
        if 'specs' in self.metadata:
            specs = self.metadata['specs']
            for i, value in enumerate(values):
                if value is not None:
                    new_values.append(value)
                else:
                    if names[i] in specs:
                        new_values.append(specs[names[i]].default)
                    else:
                        new_values.append(0.0)
        else:
            new_values = [x if x is not None else 0.0 for x in values]
        return new_values

    # // Allow incremental building of controls.
    # BUG, BUG: de cada parámetro value hace value.copy, ver posibles consecuencias...
    def _add_non_control(self, name, values):  # Not used in the standard library.
        self._add_control_name(iou.ControlName(name, None, 'noncontrol',
            values, len(self._control_names)))

    def _add_ir(self, name, values):  # *** VER dice VALUES en plural, pero salvo que se pase un array como valor todos los que calcula son escalares u objetos no iterables.
        self._add_control_name(iou.ControlName(name, len(self._controls), 'scalar',
            values, len(self._control_names)))  # values *** VER el argumento de ControlName es defaultValue que puede ser un array para expansión multicanal de controles, pero eso puede pasar acá saliendo de los argumentos?

    def _add_tr(self, name, values):
        self._add_control_name(iou.ControlName(name, len(self._controls), 'trigger',
            values, len(self._control_names)))

    def _add_ar(self, name, values):
        self._add_control_name(iou.ControlName(name, len(self._controls), 'audio',
            values, len(self._control_names)))

    def _add_kr(self, name, values, lags):  # Acá también dice lags en plural pero es un valor simple como string (symbol) o number según interpreto del código anterior.
        self._add_control_name(iou.ControlName(name, len(self._controls), 'control',
            values, len(self._control_names), lags))

    def _add_control_name(self, cn):
        self._control_names.append(cn)
        self._all_control_names.append(cn)

    def _build_controls(self):
        nn_cns = [x for x in self._control_names if x.rate == 'noncontrol']
        ir_cns = [x for x in self._control_names if x.rate == 'scalar']
        tr_cns = [x for x in self._control_names if x.rate == 'trigger']
        ar_cns = [x for x in self._control_names if x.rate == 'audio']
        kr_cns = [x for x in self._control_names if x.rate == 'control']

        arguments = [0] * len(self._control_names)
        values = []
        index = None
        ctrl_ugens = None
        lags = None
        valsize = None

        for cn in nn_cns:
            arguments[cn.arg_num] = cn.default_value

        def build_ita_controls(ita_cns, ctrl_class, method):
            nonlocal arguments, values, index, ctrl_ugens
            if ita_cns:
                values = []
                for cn in ita_cns:
                    values.append(cn.default_value)
                index = self._control_index
                ctrl_ugens = getattr(ctrl_class, method)(utl.flat(values))
                ctrl_ugens = utl.as_list(ctrl_ugens)
                ctrl_ugens = utl.reshape_like(ctrl_ugens, values) # .reshapeLike(values);
                for i, cn in enumerate(ita_cns):
                    cn.index = index
                    index += len(utl.as_list(cn.default_value))
                    arguments[cn.arg_num] = ctrl_ugens[i]
                    self._set_control_names(ctrl_ugens[i], cn)

        build_ita_controls(ir_cns, iou.Control, 'ir')
        build_ita_controls(tr_cns, iou.TrigControl, 'kr')
        build_ita_controls(ar_cns, iou.AudioControl, 'ar')

        if kr_cns:
            values = []
            lags = []
            for cn in kr_cns:
                values.append(cn.default_value)
                valsize = len(utl.as_list(cn.default_value))
                if valsize > 1:
                    lags.extend(utl.wrap_extend(utl.as_list(cn.lag), valsize))
                else:
                    lags.append(cn.lag)
            index = self._control_index

            if any(x != 0 for x in lags):
                ctrl_ugens = iou.LagControl.kr(utl.flat(values), lags)  # LagControl.kr(values.flat, lags) //.asArray.reshapeLike(values);
            else:
                ctrl_ugens = iou.Control.kr(utl.flat(values))  # Control.kr(values.flat)
            ctrl_ugens = utl.as_list(ctrl_ugens) # .asArray
            ctrl_ugens = utl.reshape_like(ctrl_ugens, values)  # .reshapeLike(values);

            for i, cn in enumerate(kr_cns):
                cn.index = index
                index += len(utl.as_list(cn.default_value))
                arguments[cn.arg_num] = ctrl_ugens[i]
                self._set_control_names(ctrl_ugens[i], cn)

        self._control_names = [
            x for x in self._control_names if x.rate != 'noncontrol']
        return arguments

    def _set_control_names(self, ctrl_ugens, cn):
        if isinstance(ctrl_ugens, list):
            for ctrl_ugen in ctrl_ugens:
                ctrl_ugen.name = cn.name
        else:
            ctrl_ugens.name = cn.name

    def _finish_build(self):
        self._add_copies_if_needed()  # ping, only for WidthFirstUGen ugens.
        self._optimize_graph()
        self._collect_constants()
        self._check_inputs()  # // Will die on error.
        # // re-sort graph. reindex.
        self._topological_sort()
        self._index_ugens()
        # UGen.buildSynthDef = nil; moved to SynthDef in _build try/except

    def _add_copies_if_needed(self):
        # // Could also have PV_UGens store themselves in a separate collection.
        for child in self._width_first_ugens:
            if isinstance(child, ffu.PV_ChainUGen):
                child._add_copies_if_needed()  # pong


    # OC: Multi channel expansion causes a non optimal breadth-wise
    # ordering of the graph. The topological sort below follows
    # branches in a depth first order, so that cache performance
    # of connection buffers is optimized.

    def _optimize_graph(self):  # ping
        self._init_topo_sort()

        self._rewrite_in_progress = True
        for ugen in self._children[:]:  # ***** Hace _children.copy.do porque modifica los valores de la lista sobre la que itera. VER RECURSIVIDAD: SI MODIFICA UN VALOR ACCEDIDO POSTERIORMENTE None._optimize_graph FALLA??
            ugen._optimize_graph()  # pong
        self._rewrite_in_progress = False

        # // Fixup removed ugens.
        old_size = len(self._children)
        self._children = [x for x in self._children if x is not None]
        if old_size != len(self._children):
            self._index_ugens()

    def _init_topo_sort(self):  # ping
        self._available = []
        for ugen in self._children:
            ugen._antecedents = set()
            ugen._descendants = set()
        for ugen in self._children:
            # // This populates the _descendants and _antecedents.
            ugen._init_topo_sort()  # pong
        for ugen in reversed(self._children):
            # // All ugens with no antecedents are made available.
            ugen._make_available()

    def _index_ugens(self):
        for i, ugen in enumerate(self._children):
            ugen._synth_index = i

    def _collect_constants(self):  # ping
        for ugen in self._children:
            ugen._collect_constants()  # pong

    def _check_inputs(self):  # ping
        first_err = None
        for ugen in self._children:
            err = ugen._check_inputs() # pong, en sclang devuelve nil o un string, creo que esos serían todos los casos según la lógica de este bloque.
            if err: # *** TODO EN SCLANG ES ASIGNA A err Y COMPRUEBA notNil, acá puede ser none, pero ver qué retornan de manera sistemática, ver return acá abajo.
                # err = ugen.class.asString + err;
                # err.postln;
                # ugen._dump_args
                if first_err is None:
                    first_err = ugen.name + ' ' + err
        if first_err:
            #"SynthDef % build failed".format(this.name).postln;
            raise ValueError(first_err)
        return True # porque ugen._check_inputs() retorna nil y acá true

    def _topological_sort(self):
        self._init_topo_sort()
        ugen = None
        out_stack = []
        while len(self._available) > 0:
            ugen = self._available.pop()
            ugen._arrange(out_stack)
        self._children = out_stack
        self._cleanup_topo_sort()

    def _cleanup_topo_sort(self):
        for ugen in self._children:
            ugen._antecedents = set()
            ugen._descendants = set()
            ugen._width_first_antecedents = []


    # // UGens do these (ping pong methods).

    def _add_ugen(self, ugen):
        if not self._rewrite_in_progress:
            ugen._synth_index = len(self._children)
            ugen._width_first_antecedents = self._width_first_ugens[:]
            self._children.append(ugen)

    def _remove_ugen(self, ugen):
        # // Lazy removal: clear entry and later remove all None entries.
        self._children[ugen._synth_index] = None

    def _replace_ugen(self, a, b):
        if not isinstance(b, ugn.SynthObject):
            raise Exception('_replace_ugen assumes a SynthObject')

        b._width_first_antecedents = a._width_first_antecedents
        b._descendants = a._descendants
        b._synth_index = a._synth_index
        self._children[a._synth_index] = b

        for item in self._children:
            if item is not None:
                for i, input in enumerate(item.inputs):
                    if input is a:
                        aux = list(item.inputs)
                        aux[i] = b
                        item._inputs = tuple(aux)

    def _add_constant(self, value):
        if value not in self._constant_set:
            self._constant_set.add(value)
            self._constants[value] = len(self._constants)

    def dump_ugens(self):
        """Print the generated graph instructions in a human readable way."""

        print(self._name)
        for ugen in self._children:
            inputs = None
            if ugen.inputs is not None:
                inputs = [
                    x._dump_name() if isinstance(x, ugn.SynthObject)
                    else x for x in ugen.inputs]
            print([ugen._dump_name(), ugen.rate, inputs])

    def add(self, libname=None, completion_msg=None, keep_def=True):
        """Send the definition to the servers and keep a description.

        Adds the synthesis definition to the SynthDescLib specified by
        ``libname`` and sends it to the library's registered servers.
        All operations take place in memory (no scsyndef file is written).
        This method is used for most cases.

        Parameters
        ----------
        libname : str
            SynthDescLib library name. If not specified `'default'` will
            be used.
        completion_msg : function
            A function that receives a server as argument and return an OSC
            message. This message will be executed in the server after the
            definition is loaded.
        keep_def : bool
            A flag indicating if the function's code will be kept in the
            SynthDesc object, default value is `True`.
        """

        # // Make SynthDef available to all servers.
        desc = sdc.SynthDesc.new_from(self, keep_def)
        if libname is None:
            sdc.SynthDescLib.get_lib('default').add(desc)
            servers = set(
                s for s in srv.Server.all if s._status_watcher.has_booted)
        else:
            lib = sdc.SynthDescLib.get_lib(libname)
            lib.add(desc)
            servers = lib.servers
        for server in servers:
            self._do_send(server, fn.value(completion_msg, server))

    def _do_send(self, server, completion_msg):
        buffer = self.as_bytes()
        if len(buffer) < (65535 // 4):  # BUG: size limitation for rt safety, compare with ArrayedCollection:clumpBundles.
            server.send_msg('/d_recv', buffer, completion_msg)
        else:
            if server.is_local:
                _logger.warning(
                    f'SynthDef {self._name} too big for sending. '
                    'Retrying via synthdef file')
                self._write_def_file(plf.Platform.tmp_dir)
                server.send_msg(
                    '/d_load',
                    str(plf.Platform.tmp_dir / f'{self._name}.{self._SUFFIX}'),
                    completion_msg)
            else:
                _logger.warning(f'SynthDef {self._name} too big for sending')

    def as_bytes(self):
        """Binary format of the synthesis definition."""

        if self._bytes is None:
            stream = io.BytesIO()
            self._write_def_list([self], stream)
            self._bytes = stream.getbuffer()
        return self._bytes

    def _write_def_file(self, dir, overwrite=True, md_plugin=None):
        if not self.metadata.get('reconstructed', False):
            dir = dir or plf.Platform.synthdef_dir
            dir = pathlib.Path(dir)
            path = dir / f'{self._name}.{self._SUFFIX}'
            if overwrite or not path.exists():
                with open(path, 'wb') as file:
                    self._write_def_list([self], file)
                desc = sdc.SynthDesc.new_from(self)
                sdc.SynthDesc.populate_metadata_func(desc)
                desc.write_metadata(dir, md_plugin)
                sdc.SynthDescLib.get_lib('default').add(desc)

    @staticmethod
    def _write_def_list(lst, file):
        # This method is Collection-writeDef in sclang, is the only one
        # that creates the header. Called from as_bytes.
        file.write(b'SCgf')  # putString 'a null terminated String'
        frw.write_i32(file, 2)  # // file version
        frw.write_i16(file, len(lst))  # // number of defs in file.
        for synthdef in lst:
            synthdef._write_def(file)

    def _write_def(self, file):
        try:
            frw.write_pascal_str(file, self._name)
            self._write_constants(file)

            # // Controls have been added by the Control UGens.
            frw.write_i32(file, len(self._controls))
            for item in self._controls:
                frw.write_f32(file, item)

            allcns_tmp = [
                x for x in self._all_control_names if x.rate != 'noncontrol']
            frw.write_i32(file, len(allcns_tmp))
            for item in allcns_tmp:
                # comprueba if (item.name.notNil) # TODO: posible BUG? (ver arriba _set_control_names). Pero no debería poder agregarse items sin no son ControlNames. Arrays anidados como argumentos, de más de un nivel, no están soportados porque fallar _set_control_names según analicé.
                #if item.name: # TODO: y acá solo comprueba que sea un string no vacío, pero no comprueba el typo ni de name ni de item.
                if not isinstance(item, iou.ControlName): # TODO: test para debugear luego.
                    raise Exception(
                        'SynthDef self._all_control_names '
                        'has non ControlName object')
                elif not item.name: # ídem.
                    raise Exception(
                        'SynthDef self._all_control_names has '
                        f'empty ControlName object = {item.name}')
                frw.write_pascal_str(file, item.name)
                frw.write_i32(file, item.index)

            frw.write_i32(file, len(self._children))
            for item in self._children:
                item._write_def(file)

            frw.write_i16(file, len(self._variants))
            if len(self._variants) > 0:
                allcns_map = dict()
                for cn in allcns_tmp:
                    allcns_map[cn.name] = cn

                for varname, pairs in self._variants.items():
                    varname = self._name + '.' + varname
                    if len(varname) > 32:
                        _logger.warning(
                            f"variant '{varname}' name too log, "
                            "not writing more variants")
                        return False

                    varcontrols = self._controls[:]
                    for cname, values in pairs.items():
                        if allcns_map.keys().isdisjoint([cname]):
                            _logger.warning(
                                f"control '{cname}' of variant '{varname}' "
                                "not found, not writing more variants")
                            return False

                        cn = allcns_map[cname]
                        values = utl.as_list(values)
                        if len(values) > len(utl.as_list(cn.default_value)):
                            _logger.warning(
                                f"control: '{cname}' of variant: '{varname}' "
                                "size mismatch, not writing more variants")
                            return False

                        index = cn.index
                        for i, val in enumerate(values):
                            varcontrols[index + i] = val

                    frw.write_pascal_str(file, varname)
                    for item in varcontrols:
                        frw.write_f32(file, item)
            return True
        except Exception as e:
            raise Exception('SynthDef: could not write def') from e

    def _write_constants(self, file):
        size = len(self._constants)
        arr = [None] * size
        for value, index in self._constants.items():
            arr[index] = value
        frw.write_i32(file, size)
        for item in arr:
            frw.write_f32(file, item)

    # writeOnce, removed, see documentation.
    # removeAt, discarded, use SynthDescLib directly.


    # // Methods for special optimizations.

    def send(self, server=None, completion_msg=None):
        """Send the definition to the server.

        Parameters
        ----------
        server : Server | list
            A single server object or list.
        completion_msg : function
            A function that receives a server as argument and return an OSC
            message. This message will be executed in the server after the
            definition is loaded.
        """

        # // Only send to servers.
        if server is None:
            servers = list(
                s for s in srv.Server.all if s._status_watcher.has_booted)
        else:
            servers = utl.as_list(server)
        for server in servers:
            if not server._status_watcher.has_booted:
                _logger.warning(
                    f"Server '{server.name}' not running, "  # *** BUG in sclang: prints server.name instead of each.name
                    "could not send SynthDef")
            if self.metadata.get('reconstructed', False):
                self._load_reconstructed(
                    server, fn.value(completion_msg, server))
            else:
                self._do_send(server, fn.value(completion_msg, server))

    def _load_reconstructed(self, server, completion_msg):
        # // This method warns and does not halt because
        # // loading existing def from disk is a viable
        # // alternative to get the synthdef to the server.
        _logger.warning(
            f"SynthDef '{self._name}' was reconstructed from a "
            f"{self._SUFFIX} file, it does not contain all the "
            "required structure to send back to the server")
        if server.is_local:
            _logger.warning(f"loading from disk instead for Server '{server}'")
            bundle = ['/d_load', self.metadata['load_path'], completion_msg]
            server.send_bundle(None, bundle)
        else:
            raise Exception(
                f"Server '{server}' is remote, cannot load from disk")

    def load(self, server, completion_msg=None, dir=None):
        """Write the definition to a file that is loaded from the server.

        This method is used for definitions too large to be sent over UDP.

        Parameters
        ----------
        server : Server | list
            A single server object or list.
        completion_msg : function
            A function that receives a server as argument and return an OSC
            message. This message will be executed in the server after the
            definition is loaded.
        dir : str | pathlib.Path
            Directory in which the file is saved, if not specified
            platform's default directory is used.
        """

        server = server or srv.Server.default
        completion_msg = fn.value(completion_msg, server)
        if self.metadata.get('reconstructed', False):
            self._load_reconstructed(server, completion_msg)
        else:
            # // Should remember what dir synthDef was written to.
            dir = dir or plf.Platform.synthdef_dir
            dir = pathlib.Path(dir)
            self._write_def_file(dir)
            server.send_msg(
                '/d_load', str(dir / f'{self._name}.{self._SUFFIX}'),
                completion_msg)

    def store(self, libname='default', dir=None, completion_msg=None,
              md_plugin=None):
        """Add a description, write the definition to disk and send
        it to the registered servers.

        Similar to ``add`` but write to disk.

        Parameters
        ----------
        libname : str
            SynthDescLib library name. If not specified `'default'` will
            be used.
        dir : str | pathlib.Path
            Directory in which the file is saved, if not specified
            platform's default directory is used.
        completion_msg : function
            A function that receives a server as argument and return an OSC
            message. This message will be executed in the server after the
            definition is loaded.
        md_plugin :
            TODO: Not defined yet.
        """

        # // Write to file and make synth description.
        lib = sdc.SynthDescLib.get_lib(libname)
        dir = dir or plf.Platform.synthdef_dir
        dir = pathlib.Path(dir)
        path = dir / f'{self._name}.{self._SUFFIX}'
        if not self.metadata.get('reconstructed', False):
            with open(path, 'wb') as file:
                self._write_def_list([self], file)
            desc = sdc.SynthDesc.new_from(self)
            desc.metadata = self.metadata
            sdc.SynthDesc.populate_metadata_func(desc)
            desc.write_metadata(dir, md_plugin)
            lib.add(desc)
            for server in lib.servers:
                self._do_send(server, fn.value(completion_msg, server))
        else:
            lib.read(path)
            for server in lib.servers:
                self._load_reconstructed(
                    server, fn.value(completion_msg, server))

    # def store_once(self, libname='default', dir=None, completion_msg=None,
    #                md_plugin=None):
    #     # // This method needs a reconsideration.
    #     dir = dir or plf.Platform.synthdef_dir
    #     dir = pathlib.Path(dir)
    #     path = dir / f'{self._name}.{self._SUFFIX}'
    #     if not path.exists():
    #         self.store(libname, dir, completion_msg, md_plugin)
    #     else:
    #         # // Load synthdesc from disk because
    #         # // SynthDescLib still needs to have the info.
    #         lib = sdc.SynthDescLib.get_lib(libname)
    #         lib.read(path)

    # play
    # def store_args(self):

    # canFreeSynth.sc Is an added interface used at least by JITlib and wslib.
    # It adds too much to core, better to find another non instrusive way.
    # can_release_synth Is used by GraphBuilder.sc for automatic outputs
    # creation (e.g. in {}.play) with *wrapOut in Function-asSynthDef.
    # hasGateControl Used in canReleaseSynth.

    def __call__(self, *args, target=None, add_action='addToHead',
                 register=False, **kwargs):
        # NOTE: kwargs can duplicate args.
        arg_list = [v for p in zip(self._callable_args, args) for v in p]
        arg_list += [v for p in kwargs.items() for v in p]
        with gpp.node_param(target)._as_target().server.bind():
            return nod.Synth(self.name, arg_list, target, add_action, register)


### Decorator syntax ###

def _create_synthdef(func, **kwargs):
    sdef = SynthDef(func.__name__, func, **kwargs)
    sdef.add()  # Running servers or offline patterns.
    sac.ServerBoot.add('all', lambda server: sdef.add())  # Next boot.
    return sdef

def synthdef(func=None, **kwargs):
    """Decorator function to build and add definitions.

    The name of the decorated function becomes the name of the
    definition. After instantiation the `add` method is called
    and is also set to be called at subsequent server's boot.

    It can be used with optional keyword only arguments.

    Examples
    --------
    ::

        @synthdef
        def test(freq=440, amp=0.1, pan=0, gate=1):
            sig = SinOsc(freq) * amp
            sig *= EnvGen(Env.asr(), gate)
            Out(0, Pan2(sig, pan))

        # Same as:

        def test(freq=440, amp=0.1, pan=0, gate=1):
            ...

        sd = SynthDef('test', test)
        sd.add()

        # With arguments:

        @synthdef(rates=[0.02, 0.02], variants={'low': {'freq': 110}})
        def test(freq=440, amp=0.1, pan=0, gate=1):
            ...

        # Same as:

        def test(freq=440, amp=0.1, pan=0, gate=1):
            ...

        sd = SynthDef('test', test, [0.02, 0.02], None, {'low': {'freq': 110}})
        sd.add()

    """

    if func is None:
        # action: 'load', 'send', 'store', 'add'? (needs kwargs filtering).
        return lambda func: _create_synthdef(func, **kwargs)
    else:
        return _create_synthdef(func)



"""Buffer.sc"""

import logging
import pathlib
import time
import array
import uuid

from ..base import responsedefs as rdf
from ..base import model as mdl
from ..base import functions as fn
from ..base import platform as plf
from ..base import builtins as bi
from ..base import utils as utl
from ..base import stream as stm
from ..base import clock as clk
from . import _graphparam as gpp
from . import server as srv


__all__ = ['Buffer']


_logger = logging.getLogger(__name__)


class BufferException(Exception):
    pass


class BufferAlreadyFreed(BufferException):
    def __init__(self, method=None):
        if method is not None:
            self.args = (f"'{method}' called",)


class Buffer(gpp.UGenParameter, gpp.NodeParameter):
    _server_caches = dict()

    def __init__(self, frames=1024, channels=1, server=None, bufnum=None,
                 *, alloc=True, completion_msg=None):
        super(gpp.UGenParameter, self).__init__(self)
        # // Doesn't send.
        self._server = server or srv.Server.default
        if bufnum is None:
            self._bufnum = self._server.next_buffer_number(1)
        else:
            self._bufnum = bufnum
        self._frames = frames
        self._channels = channels
        self._sample_rate = self._server._status_watcher.sample_rate
        self._path = None
        self._start_frame = None
        self._do_on_info = None
        self._cache()
        if alloc:
            self.alloc(completion_msg)

    @property
    def frames(self):
        return self._frames

    @property
    def channels(self):
        return self._channels

    @property
    def server(self):
        return self._server

    @property
    def bufnum(self):
        return self._bufnum

    @property
    def path(self):
        return self._path

    @property
    def start_frame(self):
        return self._start_frame

    @property
    def sample_rate(self):
        return self._sample_rate

    @property
    def duration(self):
        if self._frames is None or self._sample_rate is None:
            raise ValueError('duration parameters (frames/sr) not initialized')
        return self._frames / self._sample_rate


    ### Specialized Constructors ###

    @classmethod
    def new_consecutive(cls, num_bufs=1, frames=1024, channels=1,
                        server=None, bufnum=None, completion_msg=None):
        if bufnum is None:
            buf_base = server.next_buffer_number(num_bufs)
        else:
            buf_base = bufnum
        buf_list = []
        for i in range(num_bufs):
            new_buf = cls(frames, channels, server, buf_base + i, alloc=False)
            server.send_msg(
                '/b_alloc', buf_base + i, frames,
                channels, fn.value(completion_msg, new_buf, i))
            # new_buf._cache() # NOTE: __init__ llama a _cache
            buf_list.append(new_buf)
        return buf_list

    @classmethod
    def new_read(cls, path, start_frame=0, frames=-1, server=None,
                 bufnum=None, action=None):
        # // read whole file into memory for PlayBuf etc.
        # // Adds a query as a completion message.
        obj = cls(None, None, server, bufnum, alloc=False)
        obj._do_on_info = action
        # obj._cache() # NOTE: __init__ llama a _cache
        obj.alloc_read(
            path, start_frame, frames, lambda buf: ['/b_query', buf.bufnum])
        return obj

    @classmethod
    def new_read_no_update(cls, path, start_frame=0, frames=-1, server=None,
                           bufnum=None, completion_msg=None):
        obj = cls(None, None, server, bufnum, alloc=False)
        obj.alloc_read(path, start_frame, frames, completion_msg)
        return obj

    @classmethod
    def new_read_channel(cls, path, start_frame=0, frames=-1, channels=None,
                         server=None, bufnum=None, action=None):
        obj = cls(server, None, None, bufnum, alloc=False)
        obj._do_on_info = action
        obj.alloc_read_channel(
            path, start_frame, frames, channels,
            lambda buf: ['/b_query', buf.bufnum])
        return obj

    @classmethod
    def new_cue(cls, path, start_frame=0, channels=1, buffer_size=32768,
                server=None, bufnum=None, completion_msg=None):
        # // Preload a buffer for use with DiskIn
        obj = cls.new_alloc(
            buffer_size, channels, server, bufnum,
            lambda buf: buf.read_msg(
                path, start_frame, buffer_size, 0, True, completion_msg))
        return obj

    @classmethod
    def new_load_list(cls, channel_lst, server=None, action=None):  # Was new_load_collection
        # Difference: this version receives a list of channel data (lists).
        # // Transfer a collection of numbers to a buffer through a file.
        server = server or srv.Server.default
        if server.is_local:
            channel_lst = [list(array.array('f', l)) for l in channel_lst]  # Type check & cast.
            path = str(
                plf.Platform.tmp_dir / ('SC_' + uuid.uuid4().hex + '.wav'))
            sample_rate = int(server._status_watcher.sample_rate)
            try:
                # Was using SoundFile in sclang.
                _write_wave_file(path, channel_lst, sample_rate)
            except Exception as e:
                raise BufferException('failed to write data') from e

            def remove_file(buf):
                try:
                    pathlib.Path(path).unlink()
                    buf._path = None
                except:
                    _logger.warning(f'could not delete data file: {path}')
                finally:
                    fn.value(action, buf)

            return cls.new_read(server, path, action=remove_file)
        else:
            _logger.warning("cannot call 'new_load' with a non-local Server")

    @classmethod
    def new_send_list(cls, lst, channels=1, server=None, wait=-1, action=None):  # Was send_collection
        # // Send a Collection to a buffer one UDP sized packet at a time.
        lst = list(array.array('f', lst))  # Type check & cast.
        buffer = cls(
            server, bi.ceil(len(lst) / channels), channels, alloc=False)
        # It was forkIfNeeded, can't be implemented in Python because
        # yield statment scope is different. The check for need was:
        # if isinstance(_libsc3.main.current_tt, Routine). Always fork here
        # is even a bit more clear, use action to sync externally.
        def send_func():
            buffer.alloc()
            yield from server.sync()
            buffer.send_list(lst, 0, wait, action)

        stm.Routine.run(send_func)
        return buffer


    def alloc(self, completion_msg=None):
        self._server.send_msg(*self.alloc_msg(completion_msg))

    def alloc_msg(self, completion_msg=None):
        return [
            '/b_alloc', self._bufnum, self._frames,
            self._channels, fn.value(completion_msg, self)]

    def alloc_read(self, path, start_frame=0, frames=-1,
                   completion_msg=None):
        self._path = path
        self._start_frame = start_frame
        self._server.send_msg(*self.alloc_read_msg(
            path, start_frame, frames, completion_msg))

    def alloc_read_msg(self, path, start_frame=0, frames=-1,
                       completion_msg=None):
        self._path = path
        self._start_frame = start_frame
        return [
            '/b_allocRead', self._bufnum, path,
            start_frame, frames, fn.value(completion_msg, self)]

    def alloc_read_channel(self, path, start_frame=0, frames=-1,
                           channels=None, completion_msg=None):
        self._path = path
        self._start_frame = start_frame
        self._server.send_msg(*self.alloc_read_channel_msg(
            path, start_frame, frames, channels, completion_msg))

    def alloc_read_channel_msg(self, path, start_frame=0, frames=-1,
                               channels=None, completion_msg=None):
        self._path = path
        self._start_frame = start_frame
        return [
            '/b_allocReadChannel', self._bufnum, path, start_frame,
            frames, *channels, fn.value(completion_msg, self)]

    def read(self, path, file_start_frame=0, frames=-1,
             buf_start_frame=0, leave_open=False, action=None):
        # self._cache() # NOTE: __init__ llama a _cache
        self._do_on_info = action
        self._server.send_msg(*self.read_msg(
            path, file_start_frame, frames, buf_start_frame,
            leave_open, lambda buf: ['/b_query', buf.bufnum]))

    def read_no_update(self, path, file_start_frame=0, frames=-1,
                       buf_start_frame=0, leave_open=False,
                       completion_msg=None):
        self._server.send_msg(*self.read_msg(
            path, file_start_frame, frames,
            buf_start_frame, leave_open, completion_msg))

    def read_msg(self, path, file_start_frame=0, frames=-1,
                 buf_start_frame=0, leave_open=False, completion_msg=None):
        self._path = path
        return [
            '/b_read', self._bufnum, path, file_start_frame, frames,
            buf_start_frame, leave_open, fn.value(completion_msg, self)]

    def read_channel(self, path, file_start_frame=0, frames=-1,
                     buf_start_frame=0, leave_open=False,
                     channels=None, action=None):
        self._do_on_info = action
        self._server.send_msg(*self.read_channel_msg(
            path, file_start_frame, frames, buf_start_frame,
            leave_open, channels, lambda buf: ['/b_query', buf.bufnum]))

    def read_channel_msg(self, path, file_start_frame=0, frames=-1,
                         buf_start_frame=0, leave_open=False, channels=None,
                         completion_msg=None):
        # // doesn't set my numChannels etc.
        self._path = path
        return [
            '/b_readChannel', self._bufnum, path, file_start_frame,
            frames, buf_start_frame, leave_open, *channels,
            fn.value(completion_msg, self)]

    def cue(self, path, start_frame=0, completion_msg=None):
        # // Preload a buffer for use with DiskIn.
        self._path = path
        self._server.send_msg(*self.cue_msg(path, start_frame, completion_msg))
        # NOTE: Doesn't call _cache()

    def cue_msg(self, path, start_frame=0, completion_msg=None):
        return [
            '/b_read', self._bufnum, path, start_frame, 0, 1,
            self._frames, fn.value(completion_msg, self)]  # NOTE: Uses 1 instead of True but not always.

    def load_list(self, channel_lst, start_frame=0, action=None):  # Was load_collection
        if self._server.is_local:
            framesize = (self._frames - start_frame) * self._channels
            # if len(channel_lst) > self._channels:
            #     # Channel mismatch is reported by Server.
            #     _logger.warning('channel_lst has more channels than buffer')
            for i, chn in enumerate(channel_lst[:]):
                if len(chn) > framesize:
                    _logger.warning(
                        f'channel_lst[{i}] larger than '
                        'available number of frames')
                channel_lst[i] = list(array.array('f', chn))  # Type check & cast.
            path = str(
                plf.Platform.tmp_dir / ('SC_' + uuid.uuid4().hex + '.wav'))
            sample_rate = int(self._server._status_watcher.sample_rate)
            try:
                # Was using SoundFile in sclang.
                _write_wave_file(path, channel_lst, sample_rate)
            except Exception as e:
                raise BufferException('failed to write data') from e

            def remove_file(buf):
                try:
                    pathlib.Path(path).unlink()
                    buf._path = None
                except:
                    _logger.warning(f'could not delete data file: {path}')
                finally:
                    fn.value(action, buf)

            return self.read(
                path, buf_start_frame=start_frame, action=remove_file)
        else:
            _logger.warning("cannot call 'load' with a non-local Server")

    def send_list(self, lst, start_frame=0, wait=-1, action=None):  # Was send_collection
        number = (int, float)
        if not isinstance(start_frame, number):
            raise TypeError('start_frame must be int of float')
        if not isinstance(wait, number):
            raise TypeError('wait must be int of float')
        lst = list(array.array('f', lst))  # Type check & cast.
        size = len(lst)
        if size > (self._frames - start_frame) * self._channels:
            _logger.warning('list larger than available number of frames')
        self._stream_list(
            lst, size, start_frame * self._channels, wait, action)

    def _stream_list(self, lst, size, start_frame=0, wait=-1, action=None):  # Was streamCollection
        def stream_func():
            # // wait = -1 allows an OSC roundtrip between packets.
            # // wait = 0 might not be safe in a high traffic situation.
            # // Maybe okay with tcp.
            max_bndl_size = 1626  # // Max size for setn under udp.
            pos = 0
            sublst = None
            while pos < size:
                sublst = lst[pos:pos+max_bndl_size]
                self._server.send_msg(
                    '/b_setn', self._bufnum, start_frame + pos,
                    len(sublst), *sublst)
                pos += max_bndl_size
                if wait >= 0:
                    yield wait
                else:
                    yield from self._server.sync()
            fn.value(action, self)

        stm.Routine.run(stream_func)

    # // these next two get the data and put it in a float array
    # // which is passed to action

    def load_to_list(self, index=0, count=-1, action=None):
        def load_fork():
            path = str(
                plf.Platform.tmp_dir / ('SC_' + uuid.uuid4().hex + '.wav'))
            self.write(path, 'wav', 'float', count, index)
            yield from self._server.sync()
            channel_lst = _read_wave_file(path)
            channel_lst = [list(l) for l in channel_lst]
            try:
                pathlib.Path(path).unlink()
            except:
                _logger.warning(f'could not delete data file: {path}')
            fn.value(action, channel_lst, self)

        stm.Routine.run(load_fork)

    def get_to_list(self, index=0, count=None, wait=0.01,
                    timeout=3, action=None):
        # // risky without wait
        if action is None:
            raise ValueError('get_to_list without action makes no sense')  # *** NOTE: es que el orden de los parámetros debería ser distinto pero es contrario a la norma.

        # BUG: some methods have an strict type check but some others don't.
        max_udp_size = 1633  # // Max size for getn under udp.
        pos = index = int(index)
        if count is None:
            count = int(self._frames * self._channels)
        array = []  # [0.0] * count
        ref_count = int(bi.roundup(count / max_udp_size))
        count += pos
        done = False

        def resp_func(msg, *_):
            nonlocal done, ref_count
            if msg[1] == self._bufnum:
                array.extend(msg[4:])
                ref_count -= 1
                if ref_count <= 0:
                    done = True
                    resp.free()  # *** BUG: clear() ???
                    fn.value(action, array)  #, self)  # *** NOTE: CHANGEG: because get & getn don't pass self to action, only data

        resp = rdf.OscFunc(resp_func, '/b_setn', self._server.addr)

        def getn_func():
            nonlocal pos
            while pos < count:
                getsize = bi.min(max_udp_size, count - pos)
                self._server.send_msg(*self.getn_msg(pos, getsize))
                pos += getsize
                if wait >= 0:
                    yield wait
                else:
                    yield from self._server.sync()

        stm.Routine.run(getn_func)

        # // Lose the responder if the network choked.
        def timeout_func():
            if not done:
                resp.free()
                _logger.warning('get_to_list failed, try increasing wait time')  # *** NOTE: timeout may also fail for long buffers.

        clk.SystemClock.sched(timeout, timeout_func)

    def write(self, path=None, header_format="aiff", sample_format="int24",
              frames=-1, start_frame=0, leave_open=False,
              completion_msg=None):
        if self._bufnum is None:
            raise BufferAlreadyFreed('write')

        if path is None:
            dir = plf.Platform.recording_dir
            timestamp = time.strftime('%Y%m%d_%H%M%S')
            path = dir / ('SC_' + timestamp + '.' + header_format)
        else:
            path = pathlib.Path(path)
            if not path.suffix:
                path = pathlib.Path(str(path) + '.' + header_format)

        self._server.send_msg(*self.write_msg(
            str(path), header_format, sample_format, frames,
            start_frame, leave_open, completion_msg))

    def write_msg(self, path, header_format="aiff", sample_format="int24",
                  frames=-1, start_frame=0, leave_open=False,
                  completion_msg=None):
        if self._bufnum is None:
            raise BufferAlreadyFreed('write_msg')
        # // Doesn't change my path.
        return ['/b_write', self._bufnum, path, header_format, sample_format,
                int(frames), int(start_frame), int(leave_open),
                fn.value(completion_msg, self)]

    def free(self, completion_msg=None):
        if self._bufnum is None:
            _logger.warning('Buffer has already been freed')
        self._server.send_msg(*self.free_msg(completion_msg))

    def free_msg(self, completion_msg=None):
        if self._bufnum is None:
            _logger.warning(
                'cannot call free_msg on a Buffer that has been freed')
        self._uncache()
        self._server._buffer_allocator.free(self._bufnum)
        msg = ['/b_free', self._bufnum, fn.value(completion_msg, self)]
        self._bufnum = self._frames = self._channels = None
        self._sample_rate = self._path = self._start_frame = None
        return msg

    @classmethod
    def free_all(cls, server=None):  # Move up?
        server = server if server is not None else srv.Server.default
        server.free_all_buffers()
        type(self)._clear_server_caches(server) # *** BUG: no hace _clear_server_caches de default si es nil en sclang.

    def zero(self, completion_msg=None):
        if self._bufnum is None:
            raise BufferAlreadyFreed('zero')
        self._server.send_msg(*self.zero_msg(completion_msg))

    def zero_msg(self, completion_msg=None):
        if self._bufnum is None:
            raise BufferAlreadyFreed('zero_msg')
        return ['/b_zero', self._bufnum, fn.value(completion_msg, self)]

    def set(self, index, value, *more_pairs):
        if self._bufnum is None:
            raise BufferAlreadyFreed('set')
        self._server.send_msg(*self.set_msg(index, value, *more_pairs))

    def set_msg(self, index, value, *more_pairs):
        if self._bufnum is None:
            raise BufferAlreadyFreed('set_msg')
        return ['/b_set', self._bufnum, index, value, *more_pairs]

    def setn(self, *args):
        if self._bufnum is None:
            raise BufferAlreadyFreed('setn')
        self._server.send_msg(*self.setn_msg(*args))

    def setn_msg(self, *args):
        if self._bufnum is None:
            raise BufferAlreadyFreed('setn_msg')
        nargs = []
        for control, values in utl.gen_cclumps(args, 2):
            if isnstance(values, list):
                nargs.extend([control, len(values), *values])
            else:
                nargs.extend([control, 1, values])
        return ['/b_setn', self._bufnum, *nargs]

    def get(self, index, action):
        if self._bufnum is None:
            raise BufferAlreadyFreed('get')

        def resp_func(msg, *_):
            # // The server replies with a message of the form:
            # // [/b_set, bufnum, index, value]. We want 'value,'
            # // which is at index 3.
            fn.value(action, msg[3])

        rdf.OscFunc(
            resp_func, '/b_set', self._server.addr,
            arg_template=[self._bufnum, index]).one_shot()

        self._server.send_msg(*self.get_msg(index))

    def get_msg(self, index):  # *** BUG: sclang, los *_msg van sin action acá.
        if self._bufnum is None:
            raise BufferAlreadyFreed('get_msg')
        return ['/b_get', self._bufnum, index]

    def getn(self, index, count, action):
        if self._bufnum is None:
            raise BufferAlreadyFreed('getn')

        def resp_func(msg, *_):
            # // The server replies with a message of the form:
            # // [/b_setn, bufnum, starting index, length, ...sample values].
            # // We want the sample values, which start at index 4.
            fn.value(action, msg[4:])

        rdf.OscFunc(
            resp_func, '/b_setn', self._server.addr,
            arg_template=[self._bufnum, index]).one_shot()

        self._server.send_msg(*self.getn_msg(index, count))

    def getn_msg(self, index, count):  # *** BUG: sclang, los *_msg van sin action acá.
        if self._bufnum is None:
            raise BufferAlreadyFreed('getn_msg')
        return ['b_getn', self._bufnum, index, count]

    def fill(self, start, frames, values):
        if self._bufnum is None:
            raise BufferAlreadyFreed('fill')
        self._server.send_msg(*self.fill_msg(start, frames, *values))

    def fill_msg(self, start, frames, *values):
        if self._bufnum is None:
            raise BufferAlreadyFreed('fill_msg')
        return ['/b_fill', self._bufnum, start, int(frames), *values]  # *** NOTE: puede enviar más values que frames, ignora el servidor, los checks de tipos y valores en estas clases son raros.


    ### Gen commands ###

    def _gen_action_responder(self, action):
        if action is not None:
            def resp_func(msg, *_):
                # // The server replies with a message of the form:
                # // [/done, /b_gen, bufnum]
                fn.value(action, self)

            rdf.OscFunc(
                resp_func, '/done', self._server.addr,
                arg_template=['/b_gen', self._bufnum]).one_shot()  # *** BUG: comprobar que filtra por arg_template (no recuerdo si está implementado así).

    def _gen_oflags(self, normalize, as_wavetable, clear_first):
        flags = (int(normalize), int(as_wavetable) * 2, int(clear_first) * 4)
        return sum(flags)

    def normalize(self, new_max=1, as_wavetable=False, action=None):
        if self._bufnum is None:
            raise BufferAlreadyFreed('normalize')
        self._gen_action_responder(action)
        self._server.send_msg(*self.normalize_msg(new_max, as_wavetable))

    def normalize_msg(self, new_max=1, as_wavetable=False):
        if self._bufnum is None:
            raise BufferAlreadyFreed('normalize_msg')
        if as_wavetable:
            format = 'wnormalize'
        else:
            format = 'normalize'
        return ['/b_gen', self._bufnum, format, new_max]  # format would be cmd?

    def gen(self, cmd, args=(), normalize=True, as_wavetable=True,
            clear_first=True, action=None):
        if self._bufnum is None:
            raise BufferAlreadyFreed('gen')
        self._gen_action_responder(action)
        self._server.send_msg(*self.gen_msg(
            cmd, args, normalize, as_wavetable, clear_first))

    def gen_msg(self, cmd, args=(), normalize=True, as_wavetable=True,
                clear_first=True):
        if self._bufnum is None:
            raise BufferAlreadyFreed('gen')
        oflags = self._gen_oflags(normalize, as_wavetable, clear_first)
        return ['/b_gen', self._bufnum, cmd, oflags, *args]

    def sine1(self, amps, normalize=True, as_wavetable=True,
              clear_first=True, action=None):
        if self._bufnum is None:
            raise BufferAlreadyFreed('sine1')
        self._gen_action_responder(action)
        self._server.send_msg(*self.sine1_msg(
            amps, normalize, as_wavetable, clear_first))

    def sine1_msg(self, amps, normalize=True, as_wavetable=True,
                  clear_first=True):
        if self._bufnum is None:
            raise BufferAlreadyFreed('sine1_msg')
        oflags = self._gen_oflags(normalize, as_wavetable, clear_first)
        return ['/b_gen', self._bufnum, 'sine1', oflags, *amps]

    def sine2(self, freqs, amps, normalize=True, as_wavetable=True,
              clear_first=True, action=None):
        if self._bufnum is None:
            raise BufferAlreadyFreed('sine2')
        self._gen_action_responder(action)
        self._server.send_msg(*self.sine2_msg(
            freqs, amps, normalize, as_wavetable, clear_first))

    def sine2_msg(self, freqs, amps, normalize=True, as_wavetable=True,
                  clear_first=True):
        if self._bufnum is None:
            raise BufferAlreadyFreed('sine2_msg')
        oflags = self._gen_oflags(normalize, as_wavetable, clear_first)
        return ['/b_gen', self._bufnum, 'sine2', oflags,
                utl.lace([freqs, amps], len(freqs) * 2)]

    def sine3(self, freqs, amps, phases, normalize=True, as_wavetable=True,
              clear_first=True, action=None):
        if self._bufnum is None:
            raise BufferAlreadyFreed('sine3')
        self._gen_action_responder(action)
        self._server.send_msg(*self.sine3_msg(
            freqs, amps, phases, normalize, as_wavetable, clear_first))

    def sine3_msg(self, freqs, amps, phases, normalize=True,
                  as_wavetable=True, clear_first=True):
        if self._bufnum is None:
            raise BufferAlreadyFreed('sine3_msg')
        oflags = self._gen_oflags(normalize, as_wavetable, clear_first)
        return ['/b_gen', self._bufnum, 'sine3', oflags,
                utl.lace([freqs, amps, phases], len(freqs) * 2)]

    def cheby(self, amps, normalize=True, as_wavetable=True,
              clear_first=True, action=None):
        if self._bufnum is None:
            raise BufferAlreadyFreed('cheby')
        self._gen_action_responder(action)
        self._server.send_msg(*self.cheby_msg(
            amps, normalize, as_wavetable, clear_first))

    def cheby_msg(self, amps, normalize=True, as_wavetable=True,
                  clear_first=True):
        if self._bufnum is None:
            raise BufferAlreadyFreed('cheby_msg')
        oflags = self._gen_oflags(normalize, as_wavetable, clear_first)
        return ['/b_gen', self._bufnum, 'cheby', oflags, *amps]

    def copy_data(self, dst_buffer, dst_start=0, start=0,
                  num_samples=-1, action=None):
        if self._bufnum is None:
            raise BufferAlreadyFreed('copy_data')

        if action is not None:
            def resp_func(msg, *_):
                # // The server replies with a message of the form:
                # // [/done, /b_gen, bufnum]
                fn.value(action, self, dst_buffer)

            rdf.OscFunc(
                resp_func, '/done', self._server.addr,
                arg_template=['/b_gen', self._bufnum]).one_shot()  # *** BUG: comprobar que filtra por arg_template (no recuerdo si está implementado así).

        self._server.send_msg(*self.copy_msg(
            dst_buffer, dst_start, start, num_samples))

    def copy_data_msg(self, dst_buffer, dst_start=0, start=0, num_samples=-1):
        if self._bufnum is None:
            raise BufferAlreadyFreed('copy_data_msg')
        return ['/b_gen', dst_buffer.bufnum, 'copy', dst_start,
                self._bufnum, start, num_samples]


    def close(self, completion_msg=None):
        # // Close a file, write header, after DiskOut usage.
        if self._bufnum is None:
            raise BufferAlreadyFreed('close')
        self._server.send_msg(*self.close_msg(completion_msg))

    def close_msg(self, completion_msg):
        if self._bufnum is None:
            raise BufferAlreadyFreed('close_msg')
        return ['/b_close', self._bufnum, fn.value(completion_msg, self)]

    def query(self, action=None):
        if self._bufnum is None:
            raise BufferAlreadyFreed('query')
        if action is None:
            def action(addr, bufnum, frames, channels, sample_rate):
                _logger.info(
                    f'bufnum: {bufnum}\n'
                    f'frames: {frames}\n'
                    f'channels: {channels}\n'
                    f'sample_rate: {sample_rate}')

        def resp_func(msg, *_):
            fn.value(action, *msg)

        rdf.OscFunc(
            resp_func, '/b_info', self._server.addr,
            arg_template=[self._bufnum]).one_shot()
        self._server.send_msg('/b_query', self._bufnum)

    def query_msg(self):
        if self._bufnum is None:
            raise BufferAlreadyFreed('query_msg')
        return ['/b_query', self._bufnum]

    def update_info(self, action):
        # // Add to the array here. That way, update will
        # // be accurate even if this buf has been freed.
        self._cache()
        self._do_on_info = action
        self._server.send_msg('/b_query', self._bufnum)


    ### Cache methods ###

    def _cache(self):
        # // cache Buffers for easy info updating
        type(self)._init_server_cache(self._server)
        type(self)._server_caches[self._server][self._bufnum] = self

    def _uncache(self):
        try:
            del type(self)._server_caches[self._server][self._bufnum]
            if len(type(self)._server_caches[self._server]) == 1:
                # // The 1 item would be the responder, if there is more than 1
                # // item then the rest are cached buffers, else we can remove.
                # // cx: Tho I don't see why its important. It will just have
                # // to be added back when the next buffer is added and the
                # // responder is removed when the server reboots.
                type(self)._clear_server_caches(self._server)
        except KeyError:
            pass

    @classmethod
    def _init_server_cache(cls, server):
        if server not in cls._server_caches:
            cls._server_caches[server] = dict()

            def resp_func(msg, *_):
                try:
                    buffer = cls._server_caches[server][msg[1]]
                    buffer._frames = msg[2]
                    buffer._channels = msg[3]
                    buffer._sample_rate = msg[4]
                    buffer._query_done()
                except KeyError:
                    pass

            resp = rdf.OscFunc(resp_func, '/b_info', server.addr)
            resp.permanent = True
            cls._server_caches[server]['responder'] = resp
            mdl.NotificationCenter.register(
                server, 'new_allocators',
                cls, lambda: cls._clear_server_caches(server)
            )

    @classmethod
    def _clear_server_caches(cls, server):
        try:
            cls._server_caches[server]['responder'].free()
            del cls._server_caches[server]
        except KeyError:
            pass

    @classmethod
    def cached_buffers_do(cls, server, func):
        if server in cls._server_caches: # NOTE: No uso try porque llama a una función arbitraria.
            for i, (bufnum, buf)\
            in enumerate(cls._server_caches[server].items()):
                if type(bufnum) is not str: # NOTE: esta comprobación hay que hacerla porque el responder está junto con los buffers, y la hago por string por si bufnum es de otro tipo numérico, para que no falle en silencio.
                    func(buf, i)

    @classmethod
    def cached_buffer_at(cls, server, bufnum):
        try:
            return cls._server_caches[server][bufnum]
        except KeyError:
            return None

    def _query_done(self):
        # // Called from Server when b_info is received.
        fn.value(self._do_on_info, self)
        self._do_on_info = None

    def __repr__(self):
        return (
            f'{type(self).__name__}({self._bufnum}, {self._frames}, '
            f'{self._channels}, {self._sample_rate}, {self._path})')


    ### UGen graph parameter interface ###

    def _as_ugen_input(self, *_):
        return self.bufnum

    def _as_audio_rate_input(self):
        raise TypeError("Buffer can't be used as audio rate input")


    ### Node parameter interface ###

    def _as_control_input(self):
        return self.bufnum


    # asBufWithValues # NOTE: se implementa acá, en Ref y en SimpleNumber pero no se usa en la librería estandar.


# Minimal utility functions to avoid depencencies.

import struct

def _wave_header(frames, channels, sample_rate):
    return (
        b'RIFF' +
        struct.pack('<L', frames + 36) +  # Remaining size
        b'WAVE' +
        # Format chunk.
        b'fmt ' +
        struct.pack('<L', 16) +  # Size
        struct.pack('<H', 3) +  # IEEE float PCM
        struct.pack('<H', channels) +
        struct.pack('<L', sample_rate) +
        struct.pack('<L', 4 * sample_rate * channels) +
        struct.pack('<H', 4 * channels) +
        struct.pack('<H', 32) +
        # Data chunk
        b'data' +
        struct.pack('<L', 4 * frames * channels)  # Size
    )

def _write_wave_file(path, data, sample_rate):
    # data is [ch1[float, float, ...], ch2[], ...]
    with open(path, 'w+b') as file:
        channels = len(data)
        fmt = '<' + 'f' * channels
        file.write(_wave_header(len(data[0]), channels, sample_rate))
        for frame in zip(*data):
            file.write(struct.pack(fmt, *frame))

def _read_wave_file(path):
    ret = []
    with open(path, 'r+b') as file:
        file.seek(22)
        channels = struct.unpack('<H', file.read(2))[0]
        fmt = '<' + 'f' * channels
        size = 4 * channels

        # b'fmt ' chunk length
        file.seek(16)
        ck_len = struct.unpack('<L', file.read(4))[0]
        file.seek(20 + ck_len)

        while file.read(4) != b'data':
            data = file.read(4)
            if not data:
                raise Exception('bad file format or algorithm')
            ck_len = struct.unpack('<L', data)[0]
            file.seek(file.tell() + ck_len)

        for _ in range(struct.unpack('<L', file.read(4))[0] // size):
            ret.append(struct.unpack(fmt, file.read(size)))
    return list(zip(*ret))



"""ResponseDefs.sc"""

from abc import ABC, abstractmethod
import inspect
import logging

from ..synth import server as srv
from . import functions as fn
from . import systemactions as sac
from . import model as mdl
from . import main as _libsc3
from . import utils as utl
from ._oscmatch import osc_rematch_pattern as _match_osc_address_pattern


__all__ = ['OscFunc']


_logger = logging.getLogger(__name__)


class AbstractMessageMatcher(ABC):
    @abstractmethod
    def __call__(self):
        pass


class AbstractDispatcher(ABC):
    all = set()  # TODO: Not really used, needs a metaclass.

    def __init__(self):
        self.registered = False

    @abstractmethod
    def add(self, func_proxy):
        '''Proxies call this to add themselves to this dispatcher;
        should register this if needed.'''
        pass

    @abstractmethod
    def remove(self, func_proxy):
        '''Proxies call this to remove themselves from this dispatcher;
        should unregister if needed.'''
        pass

    @abstractmethod
    def __call__(self):
        pass

    @abstractmethod
    def register(self):
        '''Register this dispatcher to listen for its message type.'''
        pass

    @abstractmethod
    def unregister(self):
        '''Unregister this dispatcher so it no longer listens.'''
        pass

    def free(self):
        self.unregister()
        type(self).all.remove(self)

    @abstractmethod
    def type_key(self):
        '''This method must return an str.'''
        pass


class AbstractWrappingDispatcher(AbstractDispatcher):
    # // basis for the default dispatchers
    # // uses function wrappers for matching

    def __init__(self):
        super().__init__()
        self.active = dict()
        self.wrapped_funcs = dict()

    def add(self, func_proxy):
        mdl.NotificationCenter.register(
            func_proxy, 'function', self, self.update_func_for_func_proxy)
        func = self.wrap_func(func_proxy)
        self.wrapped_funcs[func_proxy] = func
        keys = self.get_keys_for_func_proxy(func_proxy)
        for key in keys:
            try:
                self.active[key].append(func)
            except KeyError:
                self.active[key] = [func]
        if not self.registered:
            self.register()

    def remove(self, func_proxy):
        mdl.NotificationCenter.unregister(func_proxy, 'function', self)
        keys = self.get_keys_for_func_proxy(func_proxy)
        func = self.wrapped_funcs[func_proxy]
        for key in keys:
            self.active[key].remove(func)
        del self.wrapped_funcs[func_proxy]
        if len(self.active) == 0:
            self.unregister()

    def update_func_for_func_proxy(self, func_proxy):
        func = self.wrap_func(func_proxy)
        old_func = self.wrapped_funcs[func_proxy]
        self.wrapped_funcs[func_proxy] = func
        keys = self.get_keys_for_func_proxy(func_proxy)
        for key in keys:
            i = self.active[key].index(old_func)
            self.active[key][i] = func

    @abstractmethod
    def wrap_func(self, func_proxy):
        pass

    @abstractmethod
    def get_keys_for_func_proxy(self, func_proxy):
        pass

    def free(self):
        for func_proxy in self.wrapped_funcs:
            mdl.NotificationCenter.unregister(func_proxy, 'function', self)
        super().free()


class AbstractResponderFunc(ABC):
    # _all_func_proxies is set()

    def __init__(self):
        self._func = None
        self._permanent = False
        self.src_id = None
        self.enabled = False
        self.dispatcher = None

    @property
    def func(self):
        return self._func

    @func.setter
    def func(self, value):  # prFunc_
        self._func = value
        mdl.NotificationCenter.notify(self, 'function')

    @property
    def permanent(self):
        return self._permanent

    @permanent.setter
    def permanent(self, value):
        self._permanent = value
        if value and self.enabled:
            sac.CmdPeriod.remove(self.__on_cmd_period)
        else:
            sac.CmdPeriod.add(self.__on_cmd_period)

    def enable(self):
        if not self.enabled:
            if not self.permanent:
                sac.CmdPeriod.add(self.__on_cmd_period)
            self.dispatcher.add(self)
            self.enabled = True
            type(self)._all_func_proxies.add(self)

    def disable(self):
        if not self.permanent:
            sac.CmdPeriod.remove(self.__on_cmd_period)
        self.dispatcher.remove(self)
        self.enabled = False

    def __on_cmd_period(self):  # Avoid clash.
        self.free()

    def one_shot(self):
        wrapped_func = self._func

        def one_shot_func(*args):
            self.free()
            wrapped_func(*args)

        self.func = one_shot_func

    # def fix(self):  # Use oscfunc.permanent = True.
    #     self.permanent = True

    def free(self):
        cls = type(self)
        if self in cls._all_func_proxies:
            cls._all_func_proxies.remove(self)
        if self.enabled:
            self.disable()

    def clear(self):
        self.func = None

    @classmethod
    def _all_func_proxies(cls):
        result = dict()
        for func_proxy in cls._all_func_proxies:
            key = func_proxy.dispatcher.type_key()
            try:
                result[key].append(func_proxy)
            except KeyError:
                result[key] = [func_proxy]
        return result

    @classmethod
    def all_enabled(cls):
        result = dict()
        enabled_proxies = [x for x in cls._all_func_proxies if x.enabled]
        for func_proxy in enabled_proxies:
            key = func_proxy.dispatcher.type_key()
            try:
                result[key].append(func_proxy)
            except KeyError:
                result[key] = [func_proxy]
        return result

    @classmethod
    def all_disabled(cls):
        result = dict()
        disabled_proxies = [x for x in cls._all_func_proxies if not x.enabled]
        for func_proxy in disabled_proxies:
            key = func_proxy.dispatcher.type_key()
            try:
                result[key].append(func_proxy)
            except KeyError:
                result[key] = [func_proxy]
        return result


### OSC ###


class OscFuncAddrMessageMatcher(AbstractMessageMatcher):
    # // If you need to test for address func gets wrapped in this.
    def __init__(self, addr, func):
        self.addr = addr
        self.func = func

    def __call__(self, msg, time, addr, recv_port):
        if self.addr.addr == addr.addr\
        and (self.addr.port is None or self.addr.port == addr.port):  # was matchItem
            fn.value(self.func, msg, time, addr, recv_port)


class OscFuncRecvPortMessageMatcher(AbstractMessageMatcher):
    # // If you need to test for recvPort func gets wrapped in this.
    def __init__(self, recv_port, func):
        self.recv_port = recv_port
        self.func = func

    def __call__(self, msg, time, addr, recv_port):
        if self.recv_port == recv_port:
            fn.value(self.func, msg, time, addr, recv_port)


class OscFuncBothMessageMatcher(AbstractMessageMatcher):
    def __init__(self, addr, recv_port, func):
        self.addr = addr
        self.recv_port = recv_port
        self.func = func

    def __call__(self, msg, time, addr, recv_port):
        if  self.addr.addr == addr.addr\
        and (self.addr.port is None or self.addr.port == addr.port)\
        and self.recv_port == recv_port:
            fn.value(self.func, msg, time, addr, recv_port)


class OSCArgsMatcher(AbstractMessageMatcher):
    def __init__(self, arg_template, func):
        self.arg_template = utl.as_list(arg_template)
        self.func = func

    def __call__(self, msg, time, addr, recv_port):
        args = msg[1:]
        for i, item in enumerate(self.arg_template):
            if item is not None and item != args[i]:  # was matchItem.not
                return
        fn.value(self.func, msg, time, addr, recv_port)


# // The default dispatchers below store by the 'most significant'
# // message argument for fast lookup. These are for use when more
# // than just the 'most significant' argument needs to be matched.

class OSCMessageDispatcher(AbstractWrappingDispatcher):
    def wrap_func(self, func_proxy):
        func = func_proxy.func
        src_id = func_proxy.src_id
        recv_port = getattr(func_proxy, 'recv_port', None)
        arg_template = getattr(func_proxy, 'arg_template', None)
        if arg_template is not None:
            func = OSCArgsMatcher(arg_template, func)
        if src_id is not None and recv_port is not None:
            return OscFuncBothMessageMatcher(src_id, recv_port, func)
        elif src_id is not None:
            return OscFuncAddrMessageMatcher(src_id, func)
        elif recv_port is not None:
            return OscFuncRecvPortMessageMatcher(recv_port, func)
        else:
            return func

    def get_keys_for_func_proxy(self, func_proxy):
        return [func_proxy.path]

    def __call__(self, msg, time, addr, recv_port):
        if msg[0] in self.active:
            for func in self.active[msg[0]]:
                fn.value(func, msg, time, addr, recv_port)

    def register(self):
        _libsc3.main.add_osc_recv_func(self) # thisProcess.addOSCRecvFunc(this)
        self.registered = True

    def unregister(self):
        _libsc3.main.remove_osc_recv_func(self) # thisProcess.removeOSCRecvFunc(this)
        self.registered = False

    def type_key(self):
        return 'OSC unmatched'


class OSCMessagePatternDispatcher(OSCMessageDispatcher):
    def __call__(self, msg, time, addr, recv_port):
        pattern = msg[0]
        for key, funcs in self.active.items():
            if _match_osc_address_pattern(pattern, key):
                for func in funcs:
                    fn.value(func, msg, time, addr, recv_port)

    def type_key(self):
        return 'OSC matched'


class OscFunc(AbstractResponderFunc):
    _all_func_proxies = set()
    default_dispatcher = OSCMessageDispatcher()
    default_matching_dispatcher = OSCMessagePatternDispatcher()

    @classmethod
    def _trace_func_show_status(cls, msg, time, addr, recv_port):
        log = ('OSC Message Received:\n'
               f'    time: {time}\n'
               f'    address: {addr}\n'
               f'    recv_port: {recv_port}\n'
               f'    msg: {msg}')
        _logger.info(log)

    @classmethod
    def _trace_func_hide_status(cls, msg, time, addr, recv_port):
        if msg[0] == '/status.reply'\
        and any(server.addr == addr for server in srv.Server.all):
            return
        cls._trace_func_show_status(msg, time, addr, recv_port)

    _trace_func = _trace_func_show_status
    _trace_running = False

    def __init__(self, func, path, src_id=None, *, recv_port=None,
                 arg_template=None, dispatcher=None):
        super().__init__()
        if path[0] != '/':
            path = '/' + path
        self.path = path
        self.src_id = src_id
        self.recv_port = recv_port
        if recv_port is not None:
            # This whould require to create another OscUdpInterface or
            # OSCUDPServer which is something I don't want because I don't
            # know why this is important of if it works for TCP in sclang.
            # OscFuncRecvPortMessageMatcher and OscFuncBothMessageMatcher
            # are needed only by this feature.
            _libsc3.main.open_udp_port(recv_port)  # Not implemented.
        self.arg_template = arg_template
        self._func = func
        self.dispatcher = dispatcher or type(self).default_dispatcher
        self.enable()
        # type(self)._all_func_proxies.add(self)  # Called by enable() already.

    @classmethod
    def matching(cls, func, path, src_id=None,
                 recv_port=None, arg_template=None):
        return cls(
            func, path, src_id, recv_port, arg_template,
            cls.default_matching_dispatcher)

    @classmethod
    def __on_cmd_period(cls):  # Avoid clash.
        cls.trace(False)

    @classmethod
    def trace(cls, flag=True, hide_status_msg=False):
        if flag and not cls._trace_running:
            if hide_status_msg:
                cls._trace_func = cls._trace_func_hide_status
            else:
                cls._trace_func = cls._trace_func_show_status
            _libsc3.main.add_osc_recv_func(cls._trace_func)
            sac.CmdPeriod.add(cls.__on_cmd_period)
            cls._trace_running = True
        elif cls._trace_running:
            _libsc3.main.remove_osc_recv_func(cls._trace_func)
            sac.CmdPeriod.remove(cls.__on_cmd_period)
            cls._trace_running = False

    def __repr__(self):
        return (
            f'{type(self).__name__}({self.path}, {self.src_id}, '
            f'{self.recv_port}, {self.arg_template})')


# class OscDef(OscFunc):


### MIDI ###

# MIDI implementation needs some thought, mostly because there
# are many libraries and so far sc3 is dependencies free.

class MidiFuncSrcMessageMatcher(AbstractMessageMatcher):
    # // If you need to test for srcID func gets wrapped in this.
    ...


class MidiFuncChanMessageMatcher(AbstractMessageMatcher):
    # // If you need to test for chan func gets wrapped in this.
    ...


class MidiFuncChanArrayMessageMatcher(AbstractMessageMatcher):
    # // If you need to test for chanArray func gets wrapped in this.
    ...


class MidiFuncSrcMessageMatcherNV(MidiFuncSrcMessageMatcher):
    # // Version for message types which don't pass a val.
    ...


class MidiFuncSrcSysMessageMatcher(MidiFuncSrcMessageMatcher):
    # // Version for message types which don't pass a val.
    ...


class MidiFuncSrcSysMessageMatcherND(MidiFuncSrcMessageMatcher):
    ...


class MidiFuncBothMessageMatcher(AbstractMessageMatcher):
    # // If you need to test for chan and srcID func gets wrapped in this.
    ...


class MidiFuncBothCAMessageMatcher(AbstractMessageMatcher):
    # // If you need to test for chanArray and srcID func gets wrapped in this.
    ...


class MIDIValueMatcher(AbstractMessageMatcher):
    # // If you want to test for the actual message value,
    # // the func gets wrapped in this.
    ...


class MIDIMessageDispatcher(AbstractWrappingDispatcher):
    # // For 'note_on', 'note_off', 'control', 'polytouch'.
    ...


class MIDIMessageDispatcherNV(MIDIMessageDispatcher):
    # // For 'touch', 'program', 'bend'.
    ...


class MIDISysexDispatcher(MIDIMessageDispatcher):
    # // For 'sysex'.
    ...


class MIDISysDataDispatcher(MIDIMessageDispatcher):
    # // sysrt with data.
    ...

class MIDISysDataDropIndDispatcher(MIDISysDataDispatcher):
    ...


class MIDISysNoDataDispatcher(MIDISysDataDispatcher):
    ...


class MIDIMTCtoSMPTEDispatcher(MIDISysexDispatcher):
    ...


class MIDISMPTEAssembler(AbstractMessageMatcher):
    # // Thanks to nescivi for code from MTC class!!
    ...


class MidiFunc(AbstractResponderFunc):
    ...

# class MidiDef(MidiFunc):



"""ListPatterns.sc"""

import collections

from ...base import stream as stm
from ...base import builtins as bi
from .. import pattern as ptt


class ListPattern(ptt.Pattern):
    def __init__(self, lst=None, repeats=1):
        lst = list(lst)  # raises TypeError
        if len(lst) > 0:
            self.lst = lst
            self.repeats = repeats
        else:
            raise ValueError(
                f"ListPattern '{type(self).__name__}' "
                "requires a non empty collection")
        self._is_event_pattern = any(
            isinstance(x, ptt.Pattern) and x.is_event_pattern for x in lst)

    @property
    def is_event_pattern(self):
        return self._is_event_pattern

    # copy
    # storeArgs


class Pseq(ListPattern):
    def __init__(self, lst, repeats=1, offset=0):
        super().__init__(lst, repeats)
        self.offset = int(offset)

    def __embed__(self, inval):
        # if (inval.eventAt('reverse') == true, { # Not good.
        lst = self.lst
        offset = self.offset
        for _ in bi.counter(self.repeats):
            for item in lst[offset:]:
                inval = yield from stm.embed(item, inval)
            for item in lst[:offset]:
                inval = yield from stm.embed(item, inval)
        return inval

    # storeArgs # TODO


class Pser(Pseq):
    def __embed__(self, inval):
        lst = self.lst
        offset = self.offset
        size = len(lst)
        for i in bi.counter(self.repeats):
            inval = yield from stm.embed(lst[(i + offset) % size], inval)
        return inval


# class Pindex(ptt.Pattern):
#     # I don't see the difference with Pswitch, here or in sclang.
#     def __init__(self, lst, indx_pattern, repeats=1):
#         self.lst = lst
#         self.indx_pattern = indx_pattern
#         self.repeats = repeats
#
#     # storeArgs
#
#     def __embed__(self, inval):
#         lst_stream = stm.stream(self.lst)
#         indx_stream = None
#         indx_pattern = self.indx_pattern
#         lst = size = None
#         try:
#             for _ in bi.counter(self.repeats):
#                 lst = lst_stream.next(inval)  # raises StopStream
#                 size = len(lst)
#                 indx_stream = stm.stream(indx_pattern)
#                 for i in indx_stream:
#                     inval = yield from stm.embed(lst[i % size], inval)
#         except stm.StopStream:
#             pass
#         return inval


class Pswitch(ptt.Pattern):
    def __init__(self, lst, which=0):
        self.lst = lst
        self.which = which

    def __embed__(self, inval):
        lst = self.lst
        size = len(lst)
        indx_stream = stm.stream(self.which)
        indx = None
        try:
            while True:
                indx = indx_stream.next(inval)  # raises StopStream
                inval = yield from stm.embed(lst[indx % size], inval)
        except stm.StopStream:
            return inval

    # storeArgs


class Pswitch1(Pswitch):
    def __embed__(self, inval):
        stream_lst = [stm.stream(i) for i in self.lst]
        size = len(stream_lst)
        indx_stream = stm.stream(self.which)
        indx = None
        try:
            while True:
                indx = indx_stream.next(inval)
                inval = yield stream_lst[indx % size].next(inval)
        except stm.StopStream:
            return inval


class Ptuple(ListPattern):
    def __embed__(self, inval):
        for _ in bi.counter(self.repeats):
            stream_lst = [stm.stream(i) for i in self.lst]
            try:
                while True:
                    tpl = []
                    for i in stream_lst:
                        tpl.append(i.next(inval))
                    inval = yield tuple(tpl)  # real tuple or list?
            except stm.StopStream:
                pass
        return inval


class Place(Pseq):
    def __embed__(self, inval):
        lst = self.lst
        offset = self.offset
        lst = lst[offset:] + lst[:offset]
        for j in bi.counter(self.repeats):
            for item in lst:
                if isinstance(item, (list, tuple)):
                    item = item[j % len(item)]
                inval = yield from stm.embed(item, inval)
        return inval


class Placep(Pseq):  # Was Ppatlace.
    def __embed__(self, inval):
        lst = self.lst
        size = len(lst)
        offset = self.offset
        stream_lst = [stm.stream(i) for i in lst[offset:]]
        stream_lst += [stm.stream(i) for i in lst[:offset]]
        done = 0
        for _ in bi.counter(self.repeats):
            for item in stream_lst:
                try:
                    inval = yield item.next(inval)
                except stm.StopStream:
                    done += 1
            if done == size:
                break
            else:
                done = 0
        return inval


### Random ###

class Pshuffle(ListPattern):  # Pshuf
    def __embed__(self, inval):
        slist = self.lst[:]
        bi.shuffle(slist)
        for _ in bi.counter(self.repeats):
            for item in slist:
                inval = yield from stm.embed(item)
        return inval


class Prand(ListPattern):
    # repeats should be length.
    def __embed__(self, inval):
        lst = self.lst
        size = len(lst)
        for _ in bi.counter(self.repeats):
            inval = yield from stm.embed(lst[bi.rand(size)], inval)
        return inval


class Pxrand(ListPattern):
    # repeats should be length.
    def __embed__(self, inval):
        lst = self.lst
        size = len(lst)
        index = bi.rand(size)
        for _ in bi.counter(self.repeats):
            index = (index + bi.rand(size - 1) + 1) % size
            inval = yield from stm.embed(lst[index], inval)
        return inval


class Pwrand(ListPattern):
    # repeats should be length.
    def __init__(self, lst, weights=None, repeats=1):  #, *, cum_weights=None, k=1):
        super().__init__(lst, repeats)
        self.weights = weights
        # self.cum_weights = cum_weights
        # self.k = k

    def __embed__(self, inval):
        lst = self.lst
        weights = self.weights
        ilst = range(len(weights)) if weights else range(len(lst))
        indx = None
        wstream = stm.stream(weights)
        try:
            for _ in bi.counter(self.repeats):
                indx = bi.choices(ilst, wstream.next(inval))[0]  #, cum_weights=cw, k=k)
                inval = yield from stm.embed(lst[indx], inval)
        except stm.StopStream:
            pass
        return inval


class Pslide(ListPattern):
    # // 'repeats' is the number of segments.
    # // 'len' is the length of each segment.
    # // 'step' is how far to step the start of each segment from previous.
    # // 'start' is what index to start at.
    # // indexing wraps around if goes past beginning or end.
    # // step can be negative.

    def __init__(self, lst, length=3, step=1, start=0, wrap=True, repeats=1):
        super().__init__(lst, repeats)
        self.length = length
        self.step = step
        self.start = start
        self.wrap = wrap

    def __embed__(self, inval):
        lst = self.lst
        size = len(lst)
        pos = self.start
        wrap = self.wrap
        step_stream = stm.stream(self.step)
        len_stream = stm.stream(self.length)
        lval = None
        try:
            for _ in bi.counter(self.repeats):
                lval = len_stream.next(inval)  # raises StopStream
                if wrap:
                    for j in range(lval):
                        inval = yield from stm.embed(
                            lst[bi.mod(pos + j, size)], inval)
                else:
                    for j in range(lval):
                        if pos + j < size:
                            inval = yield from stm.embed(
                                lst[pos + j], inval)
                        else:
                            return inval
                pos += step_stream.next(inval)  # raises StopStream
        except stm.StopStream:
            pass
        return inval


class Pwalk(ListPattern):
    # // random walk pattern - hjh - jamshark70@gmail.com
    def __init__(self, lst, steps=None, directions=1, start=0):
        super().__init__(lst)
        self.steps = steps or Prand([-1, 1], float('inf'))
        self.directions = 1 if directions is None else directions
        self.start = start

    # storeArgs

    def __embed__(self, inval):
        lst = self.lst
        step = None
        size = len(lst)
        index = self.start
        step_stream = stm.stream(self.steps)
        direction_stream = stm.stream(self.directions)
        direction = direction_stream.next(inval)

        try:
            while True:
                step = step_stream.next(inval)  # raises StopStream
                inval = yield from stm.embed(lst[int(index)], inval)
                step *= direction
                if index + step < 0 or index + step >= size:
                    direction = direction_stream.next(inval)  # or 1  # raises StopStream
                    step = abs(step) * bi.sign(direction)
                index += step
                index %= size
        except stm.StopStream:
            return inval



### Finite State Machine ###

class Pfsm(ListPattern):
    # // Finite State Machine
    def __embed__(self, inval):
        lst = self.lst
        index = None
        max_state = ((len(lst) - 1) // 2) - 1
        for _ in bi.counter(self.repeats):
            index = 0
            while True:
                if lst[index] is None:
                    break
                index = bi.clip(bi.choice(lst[index]), 0, max_state) * 2 + 2
                item = lst[index - 1]
                if item is None:
                    break
                inval = yield from stm.embed(item, inval)
        return inval


class Pdfsm(ListPattern):
    # // Deterministic Finite State Machine
    def __init__(self, lst, start_state=0, repeats=1):
        super().__init__(lst, repeats)
        self.start_state = start_state

    def __embed__(self, inval):
        lst = self.lst
        start_state = self.start_state
        num_states = len(lst) - 1
        curr_state = sig_state = None
        sig = state = out_stream = None
        for _ in bi.counter(self.repeats):
            curr_state = start_state
            sig_state = stm.stream(lst[0])
            try:
                while True:
                    sig = sig_state.next(inval)  # raises StopStream
                    if sig is None:
                        break
                    state = lst[curr_state + 1]
                    if sig in state:
                        curr_state, out_stream = state[sig]
                    else:
                        curr_state, out_stream = state['default']
                    if curr_state is None or curr_state >= num_states:
                        break
                    inval = yield from stm.embed(out_stream, inval)
            except stm.StopStream:
                pass
        return inval
